static char rcsid[] = "@(#)$Id: tagfilter_tag.c,v 2.1 2023/12/13 16:55:32 hurtta Exp $";

/******************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.1 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI>
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 ******************************************************************/

#include "def_melib.h"
#include "tagfilter_imp.h"

DEBUG_VAR(Debug,__FILE__,"mime");

#if ANSI_C
#define E_(x) extern x;
#define S_(x) static x;
#else
#define E_(x)
#define S_(x)
#endif

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

#include "elmme-magic.h"

#define TAGFILTER_TAGS_magic		0xFD0D

struct tagfilter_tags {
    unsigned short          magic;  /* TAGFILTER_TAGS_magic */

    const char            * context; /* for debug output */
    struct tagfilter_cmd ** tags;    /* NULL terminated      */
    struct sortlist       * sorted_tags;         
};


static struct tagfilter_stack_item * tagfilter_new_stack_item
P_((struct pager_range       * range     /* INC refcount */,
    int                        pg_flags,
    struct tagfilter_tags    * nested_tags  /* static - copies pointer */
                                            /* If NULL, current_tag->nesting is used */,
    struct tagfilter_cmd    * current_tag /* INC refcount */));
static struct tagfilter_stack_item * tagfilter_new_stack_item(range,
							      pg_flags,
							      nested_tags,
							      current_tag)
     struct pager_range       * range     /* INC refcount */;
     int                        pg_flags;
     struct tagfilter_tags    * nested_tags      /* static - copies pointer */;
     struct tagfilter_cmd     * current_tag /* INC refcount */;
{
     struct tagfilter_stack_item * ret = 
	 safe_zero_alloc(sizeof (*ret));
     
     ret->range = range;
     if (ret->range)
	 inc_pager_range_refcount(ret->range);
     ret->pg_flags = pg_flags;

     if (nested_tags) {
	 if (TAGFILTER_TAGS_magic != nested_tags->magic)
	     mime_panic(__FILE__,__LINE__,"tagfilter_new_stack_item",
			"Bad magic number (tagfilter_tags)");

	 
	 ret->nested_tags = nested_tags;    /* Assumed to be static */

     } else if (current_tag) {
	 if (TAGFILTER_CMD_magic != current_tag->magic)
	     mime_panic(__FILE__,__LINE__,"tagfilter_new_stack_item",
			"Bad magic number (tagfilter_cmd)");

	 ret->nested_tags = current_tag->nesting;
     } else
	 ret->nested_tags = NULL;

     
     ret-> current_tag =  current_tag;
     if (ret-> current_tag) 
	 tagfilter_inc_cmd_refcount(ret-> current_tag);
     /* refcount is incremented also for static commands */

     
     ret->reset_pg_flags = 0;
     ret->force_eoln     = 0;
     ret->text_visible   = 1;
          
     ret->magic   =  TAGFILTER_STACK_ITEM_magic;
     return ret;
}



enum tf_tag_sort_type {
    t_tagfilter_cmd   = TAGFILTER_CMD_magic,
    t_string          = STRING_magic,
    t_string_sort     = STRING_SORT_magic
    
};


#define TF_TAG_SORT_magic		0xFD15

struct tf_tag_sort {     /* union sort_key */
    unsigned short               magic;    /* TF_TAG_SORT_magic */

    struct tagfilter_tags      * backlink;
    struct tagfilter_selection * sel_backlink;
    
    union {

	/* Use unsigned short               magic
	   field as type indicator
	*/
	unsigned short        * type;  
	struct tagfilter_cmd  * u_tagfilter_cmd;
	struct string         * u_string;
	struct string_sort    * u_string_sort;
    } u;
    
};

#define TF_CMD_DEF_magic		0xFD16

struct tf_cmd_def {
    unsigned short             magic; /* TF_CMD_DEF_magic */

    int                        pg_flags;
    enum  tagfilter_cmd_base   code;
    struct tagfilter_tags    * nesting; /* static always */

    unsigned  int   cmd_mode  : TAGFLT_CMD_bit_count;
} NULL_tf_cmd_def = {
    TF_CMD_DEF_magic,
    0,
    tfc_none,
    NULL,
    0
};

struct tagfilter_cmd * tagfilter_cmd_from_string(text,base_code,cmd_mode_flags)
     struct string * text;
     enum  tagfilter_cmd_base base_code;
     int cmd_mode_flags;
{
    struct tagfilter_cmd * ret =  safe_zero_alloc(sizeof (*ret));

    ret->magic = TAGFILTER_CMD_magic;
    ret->refcount = 1;

    ret->ascii_command = NULL;

    ret->pg_flags  = 0;
    ret->code.base = base_code;
    ret->nesting   = NULL;
    
    ret->cmd_mode = cmd_mode_flags;

    /* Increments refcount and shares string! */
    ret->string_command = new_string_sort(text);

    return ret;
}

S_(alloc_sort_item_f alloc_tf_cmd_sort_item)
static void alloc_tf_cmd_sort_item P_((union sort_item      * res,
                                       const union sort_key   key,
                                       const union sort_item_default def
                                       ));
static void alloc_tf_cmd_sort_item(res,key,def)
     union sort_item      * res;
     const union sort_key   key;
     const union sort_item_default def;  
{
    struct tf_tag_sort       * tf_key = key.tf_tag_sort;
    enum tf_tag_sort_type      tf_key_type;
    const struct tf_cmd_def  * tf_cmd_def = def.tf_cmd;
    
    if (TF_TAG_SORT_magic != tf_key->magic)
	mime_panic(__FILE__,__LINE__,"alloc_tf_cmd_sort_item",
		   "Bad magic number (tf_tag_sort)");

    if (tf_cmd_def) {

	if (TF_CMD_DEF_magic != tf_cmd_def->magic)
	    mime_panic(__FILE__,__LINE__,"alloc_tf_cmd_sort_item",
		       "Bad magic number (tf_cmd_def)");

    }

    
    tf_key_type = * (tf_key->u.type);
    res->dummy  = NULL;
    
    switch (tf_key_type) {
    case t_tagfilter_cmd:
	res->tf_cmd = tf_key->u.u_tagfilter_cmd;
	tagfilter_inc_cmd_refcount(res->tf_cmd);
	break;
    case t_string:

	res->tf_cmd = tagfilter_cmd_from_string(tf_key->u.u_string,
						tfc_html_none,0);
	
	
	if (tf_cmd_def) {

	    res->tf_cmd->pg_flags   = tf_cmd_def->pg_flags;
	    res->tf_cmd->code.base  = tf_cmd_def->code;
	    res->tf_cmd->nesting    = tf_cmd_def->nesting;
	    res->tf_cmd->cmd_mode   = tf_cmd_def->cmd_mode;
	} 
	
	break;
    case t_string_sort:
	res->tf_cmd = safe_zero_alloc(sizeof (* (res->tf_cmd)));

	res->tf_cmd->magic = TAGFILTER_CMD_magic;
	res->tf_cmd->refcount = 1;

	res->tf_cmd->ascii_command = NULL;

	res->tf_cmd->pg_flags   = 0;
	res->tf_cmd->code.base  = tfc_none;
	res->tf_cmd->nesting    = NULL;
	res->tf_cmd->cmd_mode   = 0;
	
	if (tf_cmd_def) {

	    res->tf_cmd->pg_flags   = tf_cmd_def->pg_flags;
	    res->tf_cmd->code.base  = tf_cmd_def->code;
	    res->tf_cmd->nesting    = tf_cmd_def->nesting;
	    res->tf_cmd->cmd_mode   = tf_cmd_def->cmd_mode;	    
	} 
	
	res->tf_cmd->string_command = tf_key->u.u_string_sort;
	inc_string_sort_refcount(res->tf_cmd->string_command);

	break;
    }
    
}

S_(free_sort_item_f free_tf_cmd_sort_item)
static void free_tf_cmd_sort_item P_((union sort_item * ptr));
/* Decrements refcount */
static void free_tf_cmd_sort_item(ptr)
     union sort_item      * ptr; /* Decrements refcount */
{    
    tagfilter_release_cmd(&(ptr->tf_cmd));
}
    
S_(inc_sort_item_refcount_f inc_tf_cmd_sort_item_refcount)
static void inc_tf_cmd_sort_item_refcount P_((union sort_item item));
static void inc_tf_cmd_sort_item_refcount(item)
     union sort_item item;
{
    tagfilter_inc_cmd_refcount(item.tf_cmd);
}
    
S_(sort_item_debug_name_f tf_cmd_sort_item_name)
static struct string * tf_cmd_sort_item_name P_((const union sort_item item));
static struct string * tf_cmd_sort_item_name (item)
     const union sort_item item;
{
    struct tagfilter_cmd    * tf_cmd = item.tf_cmd;
    struct string           * ret = NULL;

    if (TAGFILTER_CMD_magic != tf_cmd->magic)
	mime_panic(__FILE__,__LINE__,"tf_cmd_sort_item_name",
		   "Bad magic number (tagfilter_cmd)");

    if (tf_cmd->string_command) {
	ret =
	    give_string_from_string_sort(tf_cmd->
					 string_command);
    } else if (tf_cmd->ascii_command) {
	ret = new_string2(ASCII_SET,s2us(tf_cmd->ascii_command));

	/* Cache it */

	tf_cmd->string_command =
	    new_string_sort(ret);
    }
    
    return ret;
}

/* Return < 0 if key < item 
   Return > 0 if key > item 
*/
S_(compare_sort_key_to_item_f tf_cmd_key_cmp_item)
static int tf_cmd_key_cmp_item P_((const union sort_key key,
                                   const union sort_item item));
static int tf_cmd_key_cmp_item(key,item)
     const union sort_key key;
     const union sort_item item;
{

    struct tagfilter_cmd    * tf_cmd = item.tf_cmd;
    struct tf_tag_sort      * tf_key = key.tf_tag_sort;
    enum tf_tag_sort_type     tf_key_type;
    struct tagfilter_tags      * backlink = NULL;
    struct tagfilter_selection * sel_backlink = NULL;

    
    int ret = 0;
    int ignore_case = 0;
    
    if (TAGFILTER_CMD_magic != tf_cmd->magic)
	mime_panic(__FILE__,__LINE__,"tf_cmd_key_cmp_item",
		   "Bad magic number (tagfilter_cmd)");

    if (TF_TAG_SORT_magic != tf_key->magic)
	mime_panic(__FILE__,__LINE__,"tf_cmd_key_cmp_item",
		   "Bad magic number (tf_tag_sort)");
    
    backlink = tf_key->backlink;
    if (backlink) {
	if (TAGFILTER_TAGS_magic != backlink->magic)
	    	mime_panic(__FILE__,__LINE__,"tf_cmd_key_cmp_item",
			   "Bad magic number (tagfilter_tags)");

    }
    
    sel_backlink = tf_key->sel_backlink;
    if (sel_backlink) {
	if (TAGFILTER_SELECTION_magic != sel_backlink->magic) 
	    mime_panic(__FILE__,__LINE__,"tf_cmd_key_cmp_item",
		       "Bad magic number (tagfilter_selection)");
	
	ignore_case = ison(sel_backlink->tagflt_mode,TAGFLT_MODE_lowerascii);
    }
    
    tf_key_type = * (tf_key->u.type);

    switch (tf_key_type) {
	struct string         * u_string;
	struct tagfilter_cmd  * u_tagfilter_cmd;
	struct string_sort    * u_string_sort;
    case t_tagfilter_cmd:

	u_tagfilter_cmd = tf_key->u.u_tagfilter_cmd;

	if (TAGFILTER_CMD_magic != u_tagfilter_cmd->magic)
	    mime_panic(__FILE__,__LINE__,"tf_cmd_key_cmp_item",
		       "Bad magic number (tagfilter_cmd)");

	/* Return < 0 if key (tf_key, u_tagfilter_cmd) < item (tf_cmd)
	   Return > 0 if key (tf_key, u_tagfilter_cmd) > item (tf_cmd)
	*/
	
	if (tf_cmd->ascii_command && u_tagfilter_cmd->ascii_command) {
	    ret = ignore_case ?
		istrcmp(u_tagfilter_cmd->ascii_command,
			tf_cmd->ascii_command) :
		strcmp(u_tagfilter_cmd->ascii_command,
		       tf_cmd->ascii_command);

	    DPRINT(Debug,21,(&Debug,
			     "tf_cmd_key_cmp_item ret=%d %s <> %s (ascii compare)%s\n",
			     ret,u_tagfilter_cmd->ascii_command,tf_cmd->ascii_command,
			     ignore_case ? ", ignore case" : ", case sensitive"));
	    
	} else if (tf_cmd->ascii_command && u_tagfilter_cmd->string_command) {
	    struct string * key_string =
		give_string_from_string_sort(u_tagfilter_cmd->string_command);
	    
	    ret = string_matches_ascii(key_string,
				       s2us(tf_cmd->ascii_command),
				       ignore_case ? SMA_ignore_case : 0,
				       SMA_op_return_order);


	    DPRINT(Debug,21,(&Debug,
			     "tf_cmd_key_cmp_item ret=%d %S <> %s%s\n",
			     ret,key_string,tf_cmd->ascii_command,
			     ignore_case ? ", ignore case" : ", case sensitive"));			    

	   	    
	    free_string(& key_string);	    
	} else if (tf_cmd->string_command && u_tagfilter_cmd->ascii_command) {

	    struct string * item_string =
		give_string_from_string_sort(tf_cmd->string_command);

	    /* Reverse order */

	    ret = -string_matches_ascii(item_string,
					s2us(u_tagfilter_cmd->ascii_command),
					ignore_case ? SMA_ignore_case : 0,
					SMA_op_return_order);

	    DPRINT(Debug,21,(&Debug,
			     "tf_cmd_key_cmp_item ret=%d %s <> %S%s\n",
			     ret,u_tagfilter_cmd->ascii_command,
			     item_string,
			     ignore_case ? ", ignore case" : ", case sensitive"));
	    
	    free_string(& item_string);	    
	} else if (tf_cmd->string_command && u_tagfilter_cmd->string_command) {
	    /* NOTE:
	       Only ascii_command can be case insensitive 
	    */

	    ret = string_sort_cmp(u_tagfilter_cmd->string_command,
				  tf_cmd->string_command);

	    
	    DPRINT(Debug,21,(&Debug,
			     "tf_cmd_key_cmp_item ret=%d\n",ret));

	}

	break;
    case t_string:

	u_string = tf_key->u.u_string;
	
	/* Return < 0 if key (tf_key, u_string) < item (tf_cmd)
	   Return > 0 if key (tf_key, u_string) > item (tf_cmd)
	*/

	if  (tf_cmd->ascii_command) {

	    ret = string_matches_ascii(u_string,
				       s2us(tf_cmd->ascii_command),
				       ignore_case ? SMA_ignore_case : 0,
				       SMA_op_return_order);
	    
	    DPRINT(Debug,21,(&Debug,
			     "tf_cmd_key_cmp_item ret=%d %S <> %s%s\n",
			     ret,u_string,tf_cmd->ascii_command,
			     ignore_case ? ", ignore case" : ", case sensitive"));
	    				      
	} else if (tf_cmd->string_command) {
	    /* NOTE:
	       Only ascii_command can be case insensitive 
	    */
	    
	    struct string_sort * temp =
		new_string_sort(u_string);

	    ret = string_sort_cmp(temp,tf_cmd->string_command);

	    DPRINT(Debug,21,(&Debug,
			     "tf_cmd_key_cmp_item ret=%d\n",ret));
	    
	    free_string_sort(& temp);		
	}
	
	break;
    case t_string_sort:

	u_string_sort = tf_key->u.u_string_sort;

	/* Return < 0 if key (tf_key, u_string_sort) < item (tf_cmd)
	   Return > 0 if key (tf_key, u_string_sort) > item (tf_cmd)
	*/

	if  (tf_cmd->ascii_command) {
	    struct string * key_string =
		give_string_from_string_sort(u_string_sort);

	    ret = string_matches_ascii(key_string,
				       s2us(tf_cmd->ascii_command),
				       ignore_case ? SMA_ignore_case : 0,
				       SMA_op_return_order);

	    DPRINT(Debug,21,(&Debug,
			     "tf_cmd_key_cmp_item ret=%d %S <> %s%s\n",
			     ret,key_string,tf_cmd->ascii_command,
			     ignore_case ? ", ignore case" : ", case sensitive"));
	    
	    free_string(& key_string);	 
	} else if (tf_cmd->string_command) {
	    /* NOTE:
	       Only ascii_command can be case insensitive 
	    */

	    ret = string_sort_cmp(u_string_sort,
				  tf_cmd->string_command);

	    DPRINT(Debug,21,(&Debug,
			     "tf_cmd_key_cmp_item ret=%d\n",ret));
	    
	}
	
	break;
    }    

    return ret;
}
    
S_(sort_key_debug_name_f tf_cmd_key_name)
static struct string * tf_cmd_key_name P_((const union sort_key key));
static struct string * tf_cmd_key_name(key)
     const union sort_key key;
{
    struct tf_tag_sort      * tf_key = key.tf_tag_sort;
    enum tf_tag_sort_type     tf_key_type;
    struct string * ret = NULL;
    
    if (TF_TAG_SORT_magic != tf_key->magic)
	mime_panic(__FILE__,__LINE__,"tf_cmd_key_name",
		   "Bad magic number (tf_tag_sort)");

    tf_key_type = * (tf_key->u.type);

    switch (tf_key_type) {
	struct string         * u_string;
	struct tagfilter_cmd  * u_tagfilter_cmd;
	struct string_sort    * u_string_sort;
    case t_tagfilter_cmd:

	u_tagfilter_cmd = tf_key->u.u_tagfilter_cmd;

	if (TAGFILTER_CMD_magic != u_tagfilter_cmd->magic)
	    mime_panic(__FILE__,__LINE__,"tf_cmd_key_name",
		       "Bad magic number (tagfilter_cmd)");

	if (u_tagfilter_cmd->string_command) {
	    ret =
		give_string_from_string_sort(u_tagfilter_cmd->
					     string_command);
	} else if (u_tagfilter_cmd->ascii_command) {
	    ret = new_string2(ASCII_SET,
			      s2us(u_tagfilter_cmd->ascii_command));
	    
	    /* Cache it */
	    
	    u_tagfilter_cmd->string_command =
		new_string_sort(ret);
	}
	break;
    case t_string:

	u_string = tf_key->u.u_string;

	ret = dup_string(u_string);
	break;

    case t_string_sort:

	u_string_sort = tf_key->u.u_string_sort;

	ret = give_string_from_string_sort(u_string_sort);
	break;
    }

    return ret;
}

static struct sort_operation tagfilter_cmd_operation = {
    SORT_OPERATION_magic,
    alloc_tf_cmd_sort_item,
    free_tf_cmd_sort_item,
    inc_tf_cmd_sort_item_refcount,
    tf_cmd_sort_item_name,
    tf_cmd_key_cmp_item,
    tf_cmd_key_name
};

static const int max_unknown_tags = 10;  /* Remember 10 tags on report */


struct string **tagfilter_list_tags(unsupported_tags)
     struct sortlist * unsupported_tags;
{
    size_t count =  sort_list_len(unsupported_tags);

    struct string ** ret = safe_calloc(count+1,sizeof (ret[0]));
    size_t i = 0;

    for (i = 0; i < count; i++) {
	union sort_item  res;

	res.tf_cmd = NULL;
	
	get_sort_list_item(unsupported_tags,sort_list_get_normal,i,&res);

	if (res.tf_cmd) {
	    
	    if (TAGFILTER_CMD_magic != res.tf_cmd->magic)
		mime_panic(__FILE__,__LINE__,"tagfilter_list_tags",
			   "Bad magic number (tagfilter_cmd)");

	    if (res.tf_cmd->string_command)
		ret[i] =
		    give_string_from_string_sort(res.tf_cmd->
						 string_command);
	    else if (res.tf_cmd->ascii_command) 
		ret[i] = new_string2(ASCII_SET,
				     s2us(res.tf_cmd->ascii_command));
	    else
		ret[i] = format_string(FRM("( error on item %zu )"),i);
	    
	} else {

	    DPRINT(Debug,15,(&Debug,
			     "tagfilter_list_tags: tag #%zu -  no res.tf_cmd\n",
			     i));

	    ret[i] = format_string(FRM("( error on item %zu )"),i);
	}

	free_tf_cmd_sort_item(& res);	
	
    }

    ret[i] = NULL;

    return ret;
}


static void tagfilter_add_unknown_tag P_((struct tagfilter_global * counter,
					  struct tagfilter_cmd * tag,
					  struct tagfilter_selection * sel_backlink));
static void tagfilter_add_unknown_tag(counter,tag,sel_backlink)
     struct tagfilter_global * counter;
     struct tagfilter_cmd * tag;
     struct tagfilter_selection * sel_backlink;
{
    size_t count;

    struct tf_cmd_def tf_cmd_def;
    
    union sort_key            key;
    union sort_item_default   def;
    union sort_item           res;
    struct tf_tag_sort  tf_tag_sort;
    size_t res_idx    =  0;
    
     /* bzero is defined on hdrs/elm_defs.h */
    bzero(&tf_cmd_def,sizeof tf_cmd_def);
    tf_cmd_def = NULL_tf_cmd_def;
    
    if (TAGFILTER_GLOBAL_magic != counter->magic)
	mime_panic(__FILE__,__LINE__,"tagfilter_add_unknown_tag",
                   "Bad magic number (tagfilter_global)");


    if (TAGFILTER_CMD_magic != tag->magic)
	mime_panic(__FILE__,__LINE__,"tagfilter_add_unknown_tag",
		   "Bad magic number (tagfilter_cmd)");
    
    if (!counter->unsupported_tags)
	counter->unsupported_tags =
	    alloc_sort_list(&tagfilter_cmd_operation,1);

    count = sort_list_len(counter->unsupported_tags);

    tf_tag_sort.magic             = TF_TAG_SORT_magic;
    tf_tag_sort.backlink          = NULL;
    tf_tag_sort.sel_backlink      = sel_backlink;
    tf_tag_sort.u.u_tagfilter_cmd = tag;

    res.tf_cmd = NULL;

    key.tf_tag_sort = &tf_tag_sort;
    def.tf_cmd      = &tf_cmd_def;

    if (search_sort_list_item(counter->unsupported_tags,
			      count >= max_unknown_tags ?
			      sort_list_search_normal :
			      sort_list_search_create,
			      key,def,&res,&res_idx,NULL)) {
	
	if (res.tf_cmd) {
	    
	    if (TAGFILTER_CMD_magic != res.tf_cmd->magic)
		mime_panic(__FILE__,__LINE__,"tagfilter_add_unknown_tag",
			   "Bad magic number (tagfilter_cmd)");

	     DPRINT(Debug,15,(&Debug,
			     "tagfilter_add_unknown_tag: tag found from #%zu or created.\n",
			      res_idx));

	     free_tf_cmd_sort_item(& res);
	     return;
	}
	
	DPRINT(Debug,15,(&Debug,
			 "tagfilter_add_unknown_tag: tag found from #%zu -  no res.tf_cmd\n",
			 
			 res_idx));

	free_tf_cmd_sort_item(& res);
    } else {
	DPRINT(Debug,15,(&Debug,
			 "tagfilter_add_unknown_tag: tag not found\n"));	
    }

    counter->have_other = 1;
}


static struct tagfilter_cmd * tagfilter_give_unknown_tag
     P_((struct tagfilter_global * counter,
	 struct string * tag_name,
	 struct tagfilter_selection * sel_backlink,
	 enum  tagfilter_cmd_base   base_code,
	 int cmd_mode_flags));
static struct tagfilter_cmd * tagfilter_give_unknown_tag(counter,tag_name,sel_backlink,
							 base_code,
							 cmd_mode_flags)
     struct tagfilter_global * counter;
     struct string       * tag_name;
     struct tagfilter_selection * sel_backlink;
     enum  tagfilter_cmd_base   base_code;
     int cmd_mode_flags;
{
    size_t count;

    struct tf_cmd_def tf_cmd_def;
    
    union sort_key            key;
    union sort_item_default   def;
    union sort_item           res;
    struct tf_tag_sort  tf_tag_sort;
    size_t res_idx    =  0;
    
     /* bzero is defined on hdrs/elm_defs.h */
    bzero(&tf_cmd_def,sizeof tf_cmd_def);
    tf_cmd_def = NULL_tf_cmd_def;
    tf_cmd_def.code     = base_code;
    tf_cmd_def.cmd_mode = cmd_mode_flags;
    
    if (TAGFILTER_GLOBAL_magic != counter->magic)
	mime_panic(__FILE__,__LINE__,"tagfilter_give_unknown_tag",
                   "Bad magic number (tagfilter_global)");

    if (!counter->unsupported_tags)
	counter->unsupported_tags = alloc_sort_list(&tagfilter_cmd_operation,1);

    count = sort_list_len(counter->unsupported_tags);

    tf_tag_sort.magic             = TF_TAG_SORT_magic;
    tf_tag_sort.backlink          = NULL;
    tf_tag_sort.sel_backlink      = sel_backlink;
    tf_tag_sort.u.u_string        = tag_name;

    res.tf_cmd = NULL;

    key.tf_tag_sort = &tf_tag_sort;
    def.tf_cmd      = &tf_cmd_def;

    if (search_sort_list_item(counter->unsupported_tags,
			      count >= max_unknown_tags ?
			      sort_list_search_normal :
			      sort_list_search_create,
			      key,def,&res,&res_idx,NULL)) {
	
	if (res.tf_cmd) {
	    
	    if (TAGFILTER_CMD_magic != res.tf_cmd->magic)
		mime_panic(__FILE__,__LINE__,"tagfilter_give_unknown_tag",
			   "Bad magic number (tagfilter_cmd)");

	    DPRINT(Debug,15,(&Debug,
			     "tagfilter_give_unknown_tag: tag %S found from #%zu or created.\n",
			     tag_name,res_idx));
	    
	    
	    return res.tf_cmd;
	}
	
	DPRINT(Debug,15,(&Debug,
			 "tagfilter_give_unknown_tag: tag %S found from #%zu -  no res.tf_cmd\n",
			 
			 tag_name,res_idx));

	free_tf_cmd_sort_item(& res);	
    } else {
	DPRINT(Debug,15,(&Debug,
			 "tagfilter_give_unknown_tag: tag %S not found\n",
			 tag_name));	
    }
    
    counter->have_other = 1;

    return NULL;
}

static struct sortlist *  generate_sorted_tags
    P_((struct tagfilter_tags * tags,
	struct tagfilter_selection * sel_backlink));
static struct sortlist *  generate_sorted_tags(tags,sel_backlink)
     struct tagfilter_tags * tags;
     struct tagfilter_selection * sel_backlink;
{
    struct sortlist *  ret = NULL;
    size_t count = 0;
    size_t i;

    struct tf_cmd_def tf_cmd_def;  /* Not really needed
				      on this case
				   */

    size_t                res_idx = 0;
    
    if (TAGFILTER_TAGS_magic != tags->magic)
	mime_panic(__FILE__,__LINE__,"generate_sorted_tags",
		   "Bad magic number (tagfilter_tags)");

    if (TAGFILTER_SELECTION_magic != sel_backlink->magic) 
	mime_panic(__FILE__,__LINE__,"generate_sorted_tags",
		   "Bad magic number (tagfilter_selection)");

    /* bzero is defined on hdrs/elm_defs.h */
    bzero(&tf_cmd_def,sizeof tf_cmd_def);
    tf_cmd_def = NULL_tf_cmd_def;   

       
    if (tags->tags) {
	while (tags->tags[count]) {
	    
	    if (TAGFILTER_CMD_magic != tags->tags[count]->magic)
		mime_panic(__FILE__,__LINE__,"generate_sorted_tags",
			   "Bad magic number (tagfilter_cmd)");

	    count++;
	}
    }

    DPRINT(Debug,20,(&Debug,
		     "generate_sorted_tags: Adding %zu tags for %s\n",
		     count,tags->context));
    
    ret = alloc_sort_list(&tagfilter_cmd_operation,count);


    for (i = 0; i < count; i++) {

	union sort_key            key;
	union sort_item_default   def;
	union sort_item           res;
	
	struct tf_tag_sort  tf_tag_sort;
	
	if (TAGFILTER_CMD_magic != tags->tags[i]->magic)
	    mime_panic(__FILE__,__LINE__,"generate_sorted_tags",
		       "Bad magic number (tagfilter_cmd)");

	/* bzero is defined on hdrs/elm_defs.h */
	bzero(&tf_tag_sort,sizeof tf_tag_sort);

	tf_tag_sort.magic             = TF_TAG_SORT_magic;
	tf_tag_sort.backlink          = tags;
	tf_tag_sort.sel_backlink      = sel_backlink;
	tf_tag_sort.u.u_tagfilter_cmd = tags->tags[i];
	
	res.tf_cmd = NULL;
	
	key.tf_tag_sort = &tf_tag_sort;
	def.tf_cmd      = &tf_cmd_def;

	if (search_sort_list_item(ret,sort_list_insert_hint,
				  key,def,&res,&res_idx,NULL)) {
	    
	    if (res.tf_cmd) {

		if (TAGFILTER_CMD_magic != res.tf_cmd->magic)
		    mime_panic(__FILE__,__LINE__,"generate_sorted_tags",
			       "Bad magic number (tagfilter_cmd)");
		
		if (res.tf_cmd != tags->tags[i]) {
		    DPRINT(Debug,1,(&Debug,
				    "generate_sorted_tags: failed to add %zu %Q - res_idx #%zu - got different entry",
				    i,tags->tags[i]->ascii_command,
				    res_idx));
		    if (res.tf_cmd->ascii_command) {
			DPRINT(Debug,1,(&Debug," %Q",
					res.tf_cmd->ascii_command));
		    }
		    DPRINT(Debug,1,(&Debug,"\n"));
		}
		

	    } else {
		DPRINT(Debug,1,(&Debug,
				"generate_sorted_tags: failed to add %zu %Q - res_idx #%zu - no res.tf_cmd\n",
				i,tags->tags[i]->ascii_command,
				res_idx));

	    }

	    free_tf_cmd_sort_item(& res);
	    
	} else {
	    DPRINT(Debug,1,(&Debug,
			    "generate_sorted_tags: failed to add %zu %Q - res_idx #%zu\n",
			    i,tags->tags[i]->ascii_command,
			    res_idx));
	}
    }
    
    return ret;
}

struct tagfilter_global NULL_counter = {
    TAGFILTER_GLOBAL_magic,
    NULL,
    0
};

void tagfilter_global_dest(counter)
     struct tagfilter_global *counter;
{
    if (TAGFILTER_GLOBAL_magic != counter->magic)
	mime_panic(__FILE__,__LINE__,"tagfilter_global_dest",
                   "Bad magic number (tagfilter_global)");

    if (counter->unsupported_tags)
	free_sort_list(& (counter->unsupported_tags));

    counter->have_other = 0;

}

/* Increments refount ! 
   does not recurse
*/

static struct tagfilter_cmd * locate_command
          P_((struct tagfilter_tags * tags,  /* assumed static */
	      struct string   * tag_name,
	      struct tagfilter_selection * sel_backlink));
static struct tagfilter_cmd * locate_command(tags,tag_name,
					     sel_backlink)
     struct tagfilter_tags * tags;
     struct string         * tag_name;
     struct tagfilter_selection * sel_backlink;
{

    struct tf_cmd_def tf_cmd_def;  /* Not really needed
				      on this case
				   */
    
    if (TAGFILTER_TAGS_magic != tags->magic)
	mime_panic(__FILE__,__LINE__,"locate_command",
		   "Bad magic number (tagfilter_tags)");
    
    /* bzero is defined on hdrs/elm_defs.h */
    bzero(&tf_cmd_def,sizeof tf_cmd_def);
    tf_cmd_def = NULL_tf_cmd_def;
    
    tf_cmd_def.nesting = tags;   /* By default same tags 
				    assumed static
				 */    
     
     if (! tags->sorted_tags)
	 tags->sorted_tags = generate_sorted_tags(tags,sel_backlink);
     
     if (tags->sorted_tags) {
	 union sort_key            key;
	 union sort_item_default   def;
	 union sort_item           res;
	 struct tf_tag_sort  tf_tag_sort;
	 size_t res_idx    =  0;

	 DPRINT(Debug,20,(&Debug,
			  "locate_command: Searching %S tag from %s\n",
			  tag_name,tags->context));
	 
	 tf_tag_sort.magic             = TF_TAG_SORT_magic;
	 tf_tag_sort.backlink          = tags;
	 tf_tag_sort.sel_backlink      = sel_backlink;
	 tf_tag_sort.u.u_string        = tag_name;

	 res.tf_cmd = NULL;


	 key.tf_tag_sort = &tf_tag_sort;
	 def.tf_cmd      = &tf_cmd_def;

	 if (search_sort_list_item(tags->sorted_tags,
				   sort_list_search_normal,
				   key,def,&res,&res_idx,NULL)) {

	     if (res.tf_cmd) {
		 
		 if (TAGFILTER_CMD_magic != res.tf_cmd->magic)
		     mime_panic(__FILE__,__LINE__,"locate_command",
				"Bad magic number (tagfilter_cmd)");
		 
		 DPRINT(Debug,15,(&Debug,
				  "locate_command: tag %S found from #%zu (context %s)",
				  tag_name,res_idx,
				  tags->context));

		 
		 DPRINT(Debug,15,(&Debug,", code=%d",
				  res.tf_cmd->code.code));
		 
		 switch (res.tf_cmd->code.base) {
		 case tfc_unsupported:
		     DPRINT(Debug,15,(&Debug," tfc_unsupported")); break;
		 case tfc_none:
		     DPRINT(Debug,15,(&Debug," tfc_none"));break;
		 case tfc_enriched_cmd:
		 case tfc_html_cmd:
		     break;
		 }


		 DPRINT(Debug,15,(&Debug,"\n"));
		 
		 return res.tf_cmd;
	     }

	     DPRINT(Debug,15,(&Debug,
			      "locate_command: tag %S found from #%zu (context %s) -  no res.tf_cmd\n",
			      
			      tag_name,res_idx,tags->context));

	     free_tf_cmd_sort_item(& res);

	 } else {
	     DPRINT(Debug,15,(&Debug,
			      "locate_command: tag %S not found from %s\n",
			      tag_name,tags->context));
	 }
	 	 
     } else {
	 DPRINT(Debug,9,(&Debug,
			 "locate_command: no sorted_tags when searching %S from %s\n",
			 tag_name,tags->context));
     }
     
    return NULL;
}

/* non-common parts need aleady be free'ed */



static void tagfilter_free_tag_state_common  P_((struct tagfilter_tag_state **tag_state));
static void tagfilter_free_tag_state_common(tag_state)
     struct tagfilter_tag_state **tag_state;
{
    if (TAGFILTER_TAG_STATE_magic != (*tag_state)->magic)
	mime_panic(__FILE__,__LINE__,"tagfilter_free_tag_state_common",
		   "Bad magic number (tagfilter_tag_state)");


    if ((*tag_state)->u.dummy) {
	DPRINT(Debug,1,(&Debug,
			"tagfilter_free_tag_state_common: (*tag_state)->u.dummy not NULL\n"));

	/* leaks data */
	
	free ((*tag_state)->u.dummy);
	(*tag_state)->u.dummy = NULL;
    }
    
    if ((*tag_state)-> current_tag)
	tagfilter_release_cmd(& ((*tag_state)-> current_tag));

    
    (*tag_state)->magic = 0; /* Invalidate */

    free(*tag_state);
    *tag_state = NULL;
}


struct tagfilter_tag_state * tagfilter_new_tag_state(current_tag)
     struct tagfilter_cmd    * current_tag;
{
    struct tagfilter_tag_state * ret = 
	safe_zero_alloc(sizeof (*ret));
    
    ret->magic = TAGFILTER_TAG_STATE_magic;
    
    ret->current_tag = current_tag;
    if (ret->current_tag)
	tagfilter_inc_cmd_refcount(ret->current_tag);
	    
    ret->u.dummy     = NULL;

    ret->no_context  = 0;

    if (ret->current_tag) {
	if (TAGFILTER_CMD_magic != ret->current_tag->magic)
	    mime_panic(__FILE__,__LINE__,"tagfilter_new_tag_state",
		       "Bad magic number (tagfilter_cmd)");

	ret->no_context = ison(ret->current_tag->cmd_mode,
			       TAGFLT_CMD_no_context);
    }
    
    return ret;
}

/* enriched ------------------------------ */

struct tagfilter_tags text_enriched_tags = {
    TAGFILTER_TAGS_magic,

    "text/enriched tags",
    
    &(tagfilter_enriched_cmds[0]),
    NULL
};

E_(give_top_stack_item_f give_top_text_enriched)
struct tagfilter_stack_item * give_top_text_enriched
  P_((mime_t                     * body,
      struct in_state            * state_in,
      struct out_state           * state_out,
      const struct decode_opts   * decode_opt,
      charset_t                    text_charset,
      struct tagfilter_selection * tagfilter));
struct tagfilter_stack_item * give_top_text_enriched(body,state_in,state_out,
						     decode_opt,text_charset,tagfilter)
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     const struct decode_opts   * decode_opt;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    struct tagfilter_stack_item * ret = NULL;
    struct pager_range          * enriched_body_range = NULL;

    if (MIME_magic != body->magic)
	mime_panic(__FILE__,__LINE__,"give_top_text_enriched",
		   "Bad magic number (mime_t)");
	
    if (decode_opt)
	enriched_body_range =
	    state_add_simple_pager_range(state_out,
					 /* Handles quote prefix */
					 decode_opt->range,
					 PR_WORD_WRAP|PR_JOIN_LINES|
					 PR_ADD_SPACE,0,0);
    else
	enriched_body_range =
	    state_add_simple_pager_range(state_out,
					 NULL,
					 PR_WORD_WRAP|PR_JOIN_LINES|
					 PR_ADD_SPACE,0,0);

    ret =  tagfilter_new_stack_item(enriched_body_range,
				    0,
				    &text_enriched_tags,
				    NULL);

    if (TAGFILTER_STACK_ITEM_magic != ret->magic)
	mime_panic(__FILE__,__LINE__,"give_top_text_enriched",
		   "Bad magic number (tagfilter_stack_item)");
    
    free_pager_range(&enriched_body_range);
    
    return ret;
}

static  struct pager_range * give_prev_range
     P_((struct tagfilter_stack_item **stack,
	 size_t                        stack_len,
	 int                          * inherit_pg_flags_p,
	 int                          * text_visible_p
	 ));
static  struct pager_range * give_prev_range(stack,stack_len,
					     inherit_pg_flags_p,
					     text_visible_p)
     struct tagfilter_stack_item ** stack;
     size_t                         stack_len;
     int                          * inherit_pg_flags_p;
     int                          * text_visible_p;
{
    struct pager_range * prev_range = NULL;

    if (stack && stack_len > 0) {
	size_t i = stack_len;

	int is_reset = 0;
	
	while (i > 0) {
	    i--;
	    
	    if (stack[i]) {
		if (TAGFILTER_STACK_ITEM_magic != stack[i]->magic)
		    mime_panic(__FILE__,__LINE__,
			       "give_prev_range",
			       "Bad magic number (tagfilter_stack_item)");
		if (! prev_range)
		    prev_range = stack[i]->range;

		if (! is_reset && inherit_pg_flags_p)
		    *inherit_pg_flags_p |= stack[i]->pg_flags;

		if (stack[i]->reset_pg_flags)
		    is_reset = 1;
		
		if (prev_range && is_reset)
		    break;
		if (prev_range && ! inherit_pg_flags_p)
		    break;
	    }
	}

	if (TAGFILTER_STACK_ITEM_magic != stack[stack_len-1]->magic)
	    mime_panic(__FILE__,__LINE__,
		       "give_prev_range",
		       "Bad magic number (tagfilter_stack_item)");

	if (text_visible_p)
	    *text_visible_p = stack[stack_len-1]->text_visible;

    }
    
    return prev_range;
}



E_(give_stack_item_f give_item_text_enriched)
struct tagfilter_tag_state *
  give_item_text_enriched P_((struct tagfilter_stack_item **new_item,
			      struct tagfilter_tags      * tags,
			      struct string              * tag_name,
			      struct tagfilter_stack_item **stack,
			      size_t                        stack_len,
			      size_t                     * stack_pos,
			      mime_t                     * body,
			      struct in_state            * state_in,
			      struct out_state           * state_out,
			      const struct decode_opts   * decode_opt,
			      charset_t                    text_charset,
			      struct tagfilter_selection * tagfilter,
			      struct tagfilter_global    * counter));
struct tagfilter_tag_state *give_item_text_enriched(new_item,tags,
						    tag_name,
						    stack,stack_len,
						    stack_pos,
						    body,state_in,
						    state_out,
						    decode_opt,
						    text_charset,
						    tagfilter,
						    counter)
     struct tagfilter_stack_item **new_item;
     struct tagfilter_tags      * tags;
     struct string              * tag_name;
     struct tagfilter_stack_item **stack;
     size_t                        stack_len;
     size_t                     * stack_pos;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     const struct decode_opts   * decode_opt;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
     struct tagfilter_global    * counter;   
{

    struct tagfilter_cmd       * current_tag = NULL;
    struct tagfilter_tag_state * ret = NULL;
    
    int text_visible = 1;
        
    /* Allocates new  tagfilter_tag_state 
       (also NULL return possible) -- 
       gives struct tagfilter_stack_item if
       push_stack_item() needs not be used */
           
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"give_item_text_enriched",
		   "Bad magic number (tagfilter_selection)");

    if (new_item && *new_item)
	tagfilter_free_stack_item(new_item,tagfilter);

    * stack_pos = stack_len;

    if (tags) {
	struct pager_range * prev_range =
	    give_prev_range(stack,
			    stack_len,
			    NULL,
			    & text_visible
			    );
	struct pager_range * new_range = NULL;
		       	    
	if (TAGFILTER_TAGS_magic != tags->magic)
	    mime_panic(__FILE__,__LINE__,"give_item_text_enriched",
		       "Bad magic number (tagfilter_tags)");

	if (tag_name)
	    current_tag = locate_command(tags,tag_name,tagfilter);

	/* <ParaIndent> command requires <param>
	   and is not now supported
	*/
		
	if (current_tag) {

	    if (TAGFILTER_CMD_magic != current_tag->magic)
		mime_panic(__FILE__,__LINE__,"give_item_text_enriched",
			   "Bad magic number (tagfilter_cmd)");

	    if (text_visible && ison(current_tag->cmd_mode,
				     TAGFLT_CMD_hide_text))
		text_visible = 0;
	    else if (ison(current_tag->cmd_mode,
			  TAGFLT_CMD_show_text))
		text_visible = 1;
	    
	    switch (current_tag->code.enriched) {
		
	    case tfc_enriched_unsupported:

		
		tagfilter_add_unknown_tag(counter,current_tag,tagfilter);

		break;
		
	    case tfc_enriched_none:      break;
		
	    case tfc_center:

		/* Print new pager range 

		   quote level, range flags, indent are automatically
		   incremented with  prev_range

		 */
		
		new_range = state_add_simple_pager_range(state_out,
							 prev_range,
							 /* Not work very well */
							 PR_CENTER_THIS,
							 0 /* quote level */,
							 0 /* indent */);
		
		break;

	    case tfc_nojoin: {

		/* Calculate absolute range flags */
		int range_flags = PR_FLAGS_ABSOLUTE;
		
		if (prev_range)
		    setit(range_flags,get_pager_range_flags(prev_range));
		clearit(range_flags,PR_JOIN_LINES);
		clearit(range_flags,PR_ADD_SPACE);
		clearit(range_flags,PR_SAME_LINE);


		new_range = state_add_simple_pager_range(state_out,
							 prev_range,
							 range_flags,
							 0,
							 0);		    
	    }
		break;
		
	    case tfc_quote:

		/* Print new pager range 

		   quote level, range flags, indent are automatically
		   incremented with  prev_range

		 */
		
		new_range = state_add_simple_pager_range(state_out,
							 prev_range,
							 0 /* range flags */,
							 1 /* quote level */,
							 0 /* indent */);
		break;
	    }


	    if (!new_range &&
		ison(current_tag->cmd_mode,TAGFLT_CMD_force_newline)) {

		DPRINT(Debug,21,(&Debug,
				 "give_item_text_enriched: %S -- TAGFLT_CMD_force_newline set, forcing new range\n",
				 tag_name));
		
		new_range = state_add_simple_pager_range(state_out,
							 prev_range,
							 0,0,0);

	    }
	    
	    if (new_item) {

		DPRINT(Debug,21,(&Debug,
				 "give_item_text_enriched: %S pg_flags %d %s\n",
				 tag_name,current_tag->pg_flags,
				 give_pg_flags(current_tag->pg_flags)));
		
		*new_item =
		    tagfilter_new_stack_item(new_range,
					     current_tag->pg_flags,
					     tags /* PASSTRU for enriched */,
					     current_tag);

		if (TAGFILTER_STACK_ITEM_magic != (*new_item)->magic)
		    mime_panic(__FILE__,__LINE__,"give_item_text_enriched",
			       "Bad magic number (tagfilter_stack_item)");

		(*new_item)->text_visible = text_visible;
	    }
	    
	    switch (current_tag->code.enriched) {
	    case tfc_enriched_unsupported:
	    case tfc_enriched_none:        break;

		
	    case tfc_center:
		break;

	    case tfc_nojoin:
		break;
		
	    case tfc_quote:
		break;
	    }
	    
	    if (new_range)
		free_pager_range(& new_range);

	}		
    } else {
	if (tag_name) {
	    DPRINT(Debug,21,(&Debug,
			     "give_item_text_enriched: %S - no tags\n",
			     tag_name));
	}
    }

    if (!current_tag && tag_name) {
	struct pager_range * prev_range  UNUSED_VAROK =
	    give_prev_range(stack,
			    stack_len,
			    NULL,
			    & text_visible
			    );
	if (counter) 	   
	    current_tag = tagfilter_give_unknown_tag(counter,tag_name,
						     tagfilter,
						     tfc_unsupported,0
						     );

	if (!current_tag)
	    current_tag = tagfilter_cmd_from_string(tag_name,
						    tfc_unsupported,0);


	if (TAGFILTER_CMD_magic != current_tag->magic)
	    mime_panic(__FILE__,__LINE__,"give_item_text_enriched",
		       "Bad magic number (tagfilter_cmd)");
	
	if (new_item) {

	    DPRINT(Debug,21,(&Debug,
			     "give_item_text_enriched: %S pg_flags %d %s\n",
			     tag_name,current_tag->pg_flags,
			     give_pg_flags(current_tag->pg_flags)));
	    
	    *new_item =
		tagfilter_new_stack_item(NULL,
					 current_tag->pg_flags,
					 tags /* PASHTRU for enriched */,
					 current_tag);
	    
	    if (TAGFILTER_STACK_ITEM_magic != (*new_item)->magic)
		mime_panic(__FILE__,__LINE__,"give_item_text_enriched",
			   "Bad magic number (tagfilter_stack_item)");
	    
	    (*new_item)->text_visible = text_visible;
	}
    }

    if (current_tag)
	ret = tagfilter_new_tag_state(current_tag);

    if (new_item && *new_item && current_tag) {
	if (TAGFILTER_STACK_ITEM_magic != (*new_item)->magic)
	    mime_panic(__FILE__,__LINE__,"give_item_text_enriched",
		       "Bad magic number ( tagfilter_stack_item)");
	if (TAGFILTER_CMD_magic != current_tag->magic)
	    mime_panic(__FILE__,__LINE__,"give_item_text_enriched",
		       "Bad magic number (tagfilter_cmd)");

	(*new_item)->force_eoln =
	    ison(current_tag->cmd_mode,TAGFLT_CMD_force_newline);
	
    }

    if (current_tag)
	tagfilter_release_cmd(& current_tag);
    
    return ret;

}

static struct tagfilter_tag_state * tagfilter_locate_stack_item
   P_((struct tagfilter_selection * tagfilter,
       struct string_sort        ** res_tag_name_X,
       struct string               * tag_name,
       struct tagfilter_stack_item **stack,
       size_t                        stack_len,
       size_t                      * stack_pos));
static struct tagfilter_tag_state *
    tagfilter_locate_stack_item(tagfilter,res_tag_name_X,
				tag_name,stack,stack_len,stack_pos)
     struct tagfilter_selection   * tagfilter;
     struct string_sort          ** res_tag_name_X;
     struct string                * tag_name;
     struct tagfilter_stack_item ** stack;
     size_t                         stack_len;
     size_t                      * stack_pos;
{
    struct tagfilter_tag_state * ret        = NULL;
    struct string_sort         * tag_name_X = * res_tag_name_X;
    int ignore_case = 0;
    
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"tagfilter_locate_stack_item",
		   "Bad magic number (tagfilter_selection)");

    ignore_case = ison(tagfilter->tagflt_mode,TAGFLT_MODE_lowerascii);

    * stack_pos = stack_len;
    
    if (tag_name && stack && stack_len > 0) {
	size_t i = stack_len;
	
	while (i > 0) {
	    i--;
	    
	    if (stack[i]) {
		if (TAGFILTER_STACK_ITEM_magic != stack[i]->magic)
		    mime_panic(__FILE__,__LINE__,
			       "tagfilter_locate_stack_item",
			       "Bad magic number (tagfilter_stack_item)");
		
		if (stack[i]->current_tag) {
		    struct tagfilter_cmd *cmd;
		    
		    if (TAGFILTER_CMD_magic !=
			stack[i]->current_tag->magic)
			mime_panic(__FILE__,__LINE__,
				   "tagfilter_locate_stack_item",
				   "Bad magic number (tagfilter_cmd)");
		    
		    cmd = stack[i]->current_tag;
		    
		    if (cmd->ascii_command) { 
			if (string_matches_ascii(tag_name,
						 s2us(cmd->ascii_command),
						 ignore_case ? SMA_ignore_case : 0,
						 SMA_op_normal)) {
			    
			    * stack_pos = i;
			    ret = tagfilter_new_tag_state(cmd);
			    
			    break;			
			}

		    } else if (cmd->string_command) {

			/* Increments refcount and shares string! */
			if (!tag_name_X)
			    tag_name_X = new_string_sort(tag_name);
			
			if (0 == string_sort_cmp(cmd->string_command,
						 tag_name_X)) {
			    
			    * stack_pos = i;
			    ret = tagfilter_new_tag_state(cmd);
			    
			    break;			    
			}			

		    }
		}
	    }
	}
    }
    

    * res_tag_name_X = tag_name_X ;
    
    return ret;
}


E_(locate_stack_item_f locate_item_text_enriched)
struct tagfilter_tag_state *
    locate_item_text_enriched P_((struct string               * tag_name,
				  struct tagfilter_stack_item **stack,
				  size_t                        stack_len,
				  size_t                      * stack_pos
				  /* or stack_len if not found */,
				  mime_t                     * body,
				  struct in_state            * state_in,
				  struct out_state           * state_out,
				  charset_t                    text_charset,
				  struct tagfilter_selection * tagfilter));
struct tagfilter_tag_state * locate_item_text_enriched(tag_name,
						       stack,stack_len,
						       stack_pos,body,
						       state_in,state_out,
						       text_charset,tagfilter)
     struct string               * tag_name;
     struct tagfilter_stack_item **stack;
     size_t                        stack_len;
     size_t                      * stack_pos
     /* or stack_len if not found */;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    /* Allocates new  tagfilter_tag_state 
       (also NULL return possible) */
    
    struct tagfilter_tag_state * ret = NULL;

    struct string_sort  * tag_name_X = NULL;  /* Possible lowercased
						 already 
					      */
    
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"locate_item_text_enriched",
		   "Bad magic number (tagfilter_selection)");

    ret = tagfilter_locate_stack_item(tagfilter,
				      &tag_name_X,
				      tag_name,
				      stack,stack_len,stack_pos);
    
    if (tag_name_X)
	free_string_sort(& tag_name_X);
    
    
    return ret;

}

E_(end_tag_state_f end_tag_text_enriched)
void end_tag_text_enriched P_((struct tagfilter_tag_state **tag_state,
			       mime_t                     * body,
			       struct in_state            * state_in,
			       struct out_state           * state_out,
			       charset_t                    text_charset,
			       struct tagfilter_selection * tagfilter));
void end_tag_text_enriched(tag_state,body,state_in,state_out,text_charset,tagfilter)
     struct tagfilter_tag_state **tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"end_tag_text_enriched",
		   "Bad magic number (tagfilter_selection)");
    
    
    if (TAGFILTER_TAG_STATE_magic != (*tag_state)->magic)
	mime_panic(__FILE__,__LINE__,"end_tag_text_enriched",
		   "Bad magic number (tagfilter_tag_state)");
    
    /* Free enriched specific parts */
    if ((*tag_state)->u.dummy) {

	DPRINT(Debug,1,(&Debug,
			"end_tag_text_enriched: (*tag_state)->u.dummy is not NULL\n"));
    }
    tagfilter_free_tag_state_common (tag_state);
}

E_(push_stack_item_f push_tag_text_enriched)
struct tagfilter_stack_item *
   push_tag_text_enriched P_((struct tagfilter_tag_state *tag_state,
			      struct tagfilter_stack_item **stack,
			      size_t                        stack_len,
			      mime_t                     * body,
			      struct in_state            * state_in,
			      struct out_state           * state_out,
			      const struct decode_opts   * decode_opt,
			      charset_t                    text_charset,
			      struct tagfilter_selection * tagfilter));
struct tagfilter_stack_item *push_tag_text_enriched(tag_state,stack, stack_len,
						    body,state_in,state_out,
						    decode_opt,
						    text_charset,tagfilter)
     struct tagfilter_tag_state *tag_state;
     struct tagfilter_stack_item **stack;
     size_t                        stack_len;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     const struct decode_opts   * decode_opt;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"push_tag_text_enriched",
		   "Bad magic number (tagfilter_selection)");
    
    
    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"push_tag_text_enriched",
		   "Bad magic number (tagfilter_tag_state)");


    /* No implicit stack items now */

    return NULL;
}

E_(close_stack_item_f close_tag_text_enriched)
int close_tag_text_enriched P_((struct tagfilter_tag_state *tag_state,
				struct tagfilter_stack_item **stack,
				size_t                        stack_len,
				size_t                        idx,
				enum stack_close_mode         close_pos,
				mime_t                     * body,
				struct in_state            * state_in,
				struct out_state           * state_out,
				charset_t                    text_charset,
				struct tagfilter_selection * tagfilter));
int close_tag_text_enriched(tag_state,stack,stack_len,idx,close_pos,
			    body,state_in,state_out,text_charset,
			    tagfilter)
     struct tagfilter_tag_state *tag_state;
     struct tagfilter_stack_item **stack;
     size_t                        stack_len;
     size_t                        idx;
     enum stack_close_mode         close_pos;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{

    /* tagfilter_free_stack_item() is called by caller -
       return 1 (or true) if item need to be free'ed
    */

    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"close_tag_text_enriched",
		   "Bad magic number (tagfilter_selection)");
    
    
    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"close_tag_text_enriched",
		   "Bad magic number (tagfilter_tag_state)");


    /* Nothing to do */
    
    return 1;
}

E_(start_attribute_f start_atr_text_enriched)
struct  tagfilter_atr_state *
   start_atr_text_enriched P_((struct string              * atr_name,
			 struct tagfilter_tag_state * tag_state,
			 mime_t                     * body,
			 struct in_state            * state_in,
			 struct out_state           * state_out,
			 charset_t                    text_charset,
			 struct tagfilter_selection * tagfilter));
struct  tagfilter_atr_state *start_atr_text_enriched(atr_name,tag_state,
						     body,state_in,
						     state_out,
						     text_charset,
						     tagfilter)
     struct string              * atr_name;
     struct tagfilter_tag_state * tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"start_atr_text_enriched",
		   "Bad magic number (tagfilter_selection)");
        
    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"start_atr_text_enriched",
		   "Bad magic number (tagfilter_tag_state)");

    
    /* text/enriched do not have attributes */

    return NULL;
}

E_(end_attribute_f end_atr_text_enriched)
void end_atr_text_enriched P_((struct tagfilter_atr_state **atr_state,
			       struct tagfilter_tag_state * tag_state,
			       mime_t                     * body,
			       struct in_state            * state_in,
			       struct out_state           * state_out,
			       charset_t                    text_charset,
			       struct tagfilter_selection * tagfilter));
void end_atr_text_enriched(atr_state,tag_state,body,state_in,state_out,
			   text_charset,tagfilter)
     struct tagfilter_atr_state **atr_state;
     struct tagfilter_tag_state * tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{

    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"end_atr_text_enriched",
		   "Bad magic number (tagfilter_selection)");
        
    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"end_atr_text_enriched",
		   "Bad magic number (tagfilter_tag_state)");

    
    /* text/enriched do not have attributes */


    free(atr_state);
    *atr_state = NULL;
}

E_(start_end_atrvalue_f start_end_atrval_text_enriched)
void start_end_atrval_text_enriched
         P_((struct tagfilter_atr_state * atr_state,
	     enum atrvalue_start_end      start_end,
	     struct tagfilter_tag_state * tag_state,
	     mime_t                     * body,
	     struct in_state            * state_in,
	     struct out_state           * state_out,
	     charset_t                    text_charset,
	     struct tagfilter_selection * tagfilter));
void start_end_atrval_text_enriched(atr_state,start_end,tag_state,
				    body,state_in,state_out,
				    text_charset,tagfilter)
     struct tagfilter_atr_state * atr_state;
     enum atrvalue_start_end      start_end;
     struct tagfilter_tag_state * tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{

    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"start_end_atrval_text_enriched",
		   "Bad magic number (tagfilter_selection)");
        
    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"start_end_atrval_text_enriched",
		   "Bad magic number (tagfilter_tag_state)");

    
    /* text/enriched do not have attributes */

}

E_(atrvalue_seg_string_f atrvalue_seg_str_text_enriched)
void atrvalue_seg_str_text_enriched
   P_((struct tagfilter_atr_state * atr_state,
       struct string              * atr_value_segment,
       struct tagfilter_tag_state * tag_state,
       mime_t                     * body,
       struct in_state            * state_in,
       struct out_state           * state_out,
       charset_t                    text_charset,
       struct tagfilter_selection * tagfilter));
void atrvalue_seg_str_text_enriched(atr_state,atr_value_segment,
				    tag_state,body,state_in,
				    state_out,text_charset,tagfilter)
     struct tagfilter_atr_state * atr_state;
     struct string              * atr_value_segment;
     struct tagfilter_tag_state * tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{

    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"atrvalue_seg_str_text_enriched",
		   "Bad magic number (tagfilter_selection)");
        
    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"atrvalue_seg_str_text_enriched",
		   "Bad magic number (tagfilter_tag_state)");

    /* text/enriched do not have attributes */
}

E_(atrvalue_seg_ent_f atrvalue_seg_ent_text_enriched)
void atrvalue_seg_ent_text_enriched
       P_((struct tagfilter_atr_state * atr_state,
	   struct out_entity          * atr_value_segment,
	   struct tagfilter_tag_state * tag_state,
	   mime_t                     * body,
	   struct in_state            * state_in,
	   struct out_state           * state_out,
	   charset_t                    text_charset,
	   struct tagfilter_selection * tagfilter));
void atrvalue_seg_ent_text_enriched(atr_state,atr_value_segment,
				    tag_state,body,state_in,state_out,
				    text_charset,tagfilter)
     struct tagfilter_atr_state * atr_state;
     struct out_entity          * atr_value_segment;
     struct tagfilter_tag_state * tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"atrvalue_seg_ent_text_enriched",
		   "Bad magic number (tagfilter_selection)");
        
    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"atrvalue_seg_ent_text_enriched",
		   "Bad magic number (tagfilter_tag_state)");

    /* text/enriched do not have attributes */
}

E_(stack_item_clear_f stack_item_clear_text_enriched)
void stack_item_clear_text_enriched
  P_((struct tagfilter_stack_item * item,
      struct tagfilter_selection  * tagfilter));
void stack_item_clear_text_enriched(item,tagfilter)
     struct tagfilter_stack_item * item;
     struct tagfilter_selection  * tagfilter;
{
    if (TAGFILTER_STACK_ITEM_magic != item->magic)
	mime_panic(__FILE__,__LINE__,
		   "stack_item_clear_text_enriched",
		   "Bad magic number (tagfilter_stack_item)");

    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"stack_item_clear_text_enriched",
		   "Bad magic number (tagfilter_selection)");

    /* Nothing to clear */
        
}


/* html ------------------------------------ */

#define TAGFILTER_HTML_STATE_magic	0xFD19

static struct tagfilter_html_state {
    unsigned short            magic; 

    struct tagfilter_push_tag {
	struct tagfilter_cmd   * cmd;
	struct tagfilter_tags  * used_tags; /* static */
    } * push_tag;
	
    int                        push_tag_len;
    
} * tagfilter_new_html_state P_((void));
static struct tagfilter_html_state * tagfilter_new_html_state()
{
    struct tagfilter_html_state * ret =
	safe_zero_alloc(sizeof (*ret));

    ret->push_tag = NULL;
    ret->push_tag_len = 0;
    
    ret->magic = TAGFILTER_HTML_STATE_magic;
    return ret;
}

static void tagfilter_free_html_state P_((struct tagfilter_html_state **ptr));
static void tagfilter_free_html_state(ptr)
     struct tagfilter_html_state **ptr;
{
    if (TAGFILTER_HTML_STATE_magic != (*ptr)->magic) 
	mime_panic(__FILE__,__LINE__,"tagfilter_free_html_state",
		   "Bad magic number (tagfilter_selection)");

    if ((*ptr)->push_tag) {
	int i;

	for (i = 0; i < (*ptr)->push_tag_len; i++) {
	    if ((*ptr)->push_tag[i].cmd)
		tagfilter_release_cmd(& ((*ptr)->push_tag[i].cmd));
	    (*ptr)->push_tag[i].used_tags = 0;
	}

	free((*ptr)->push_tag);
	(*ptr)->push_tag = NULL;
    }
    (*ptr)->push_tag_len = 0;
    
    free(*ptr);
    *ptr = NULL;
}

static void tagfilter_html_state_push_command
  P_((struct tagfilter_html_state * html_state,
      struct tagfilter_cmd        * cmd,
      struct tagfilter_tags       * used_tags
      ));
static void tagfilter_html_state_push_command(html_state,cmd,used_tags)
     struct tagfilter_html_state * html_state;
     struct tagfilter_cmd        * cmd;
     struct tagfilter_tags       * used_tags;
{
    if (TAGFILTER_HTML_STATE_magic != html_state->magic) 
	mime_panic(__FILE__,__LINE__,
		   "tagfilter_html_state_push_command",
		   "Bad magic number (tagfilter_selection)");

    if (TAGFILTER_CMD_magic != cmd->magic)
	mime_panic(__FILE__,__LINE__,
		   "tagfilter_html_state_push_command",
		   "Bad magic number (tagfilter_cmd)");

    html_state->push_tag =
	safe_array_realloc(html_state->push_tag,
			   html_state->push_tag_len+1,
			   sizeof (html_state->push_tag[0]));

    html_state->push_tag[html_state->push_tag_len].cmd = cmd;
    tagfilter_inc_cmd_refcount(html_state->
			       push_tag[html_state->push_tag_len].cmd);
    html_state->push_tag[html_state->push_tag_len].used_tags =
	used_tags;
    
    html_state->push_tag_len++;
}

struct tagfilter_tags text_html_top = {    /* <html> .... </html>  */
    TAGFILTER_TAGS_magic,

    "text/html document root",

    &(tagfilter_html_top[0]),
    NULL
};
struct tagfilter_tags text_html_head_body = {
    TAGFILTER_TAGS_magic,

    "text/html head and body",
    
    &(tagfilter_html_head_body[0]),
    NULL
};
struct tagfilter_tags text_html_metadata = {
    TAGFILTER_TAGS_magic,

    "text/html metadata",
    
    &(tagfilter_html_metadata[0]),
    NULL
};
struct tagfilter_tags text_html_flow = {
    TAGFILTER_TAGS_magic,

    "text/html flow content",
    
    &(tagfilter_html_flow[0]),
    NULL
};

struct tagfilter_tags text_html_phrasing = {
    TAGFILTER_TAGS_magic,

    "text/html phrasing content",
    
    &(tagfilter_html_phrasing[0]),
    NULL
};


struct tagfilter_tags text_html_close_head = {
    /* Not really working */
    TAGFILTER_TAGS_magic,

    "text/html closing tags for <head>",
    
    &(tagfilter_html_close_head[0]),
    NULL
};

struct tagfilter_tags text_html_noscript_head = {
    TAGFILTER_TAGS_magic,

    "text/html content of <noscript> on <head>",
    
    &(tagfilter_html_noscript_head[0]),
    NULL
};

struct tagfilter_tags text_html_datalist = {
    TAGFILTER_TAGS_magic,

    "text/html content of <datalist>",
    &(tagfilter_html_datalist[0]),
    NULL
};

struct tagfilter_tags text_html_details = {
    TAGFILTER_TAGS_magic,
    "text/html content of <details>",
    &(tagfilter_html_details[0]),
    NULL
};

struct tagfilter_tags text_html_dl = {
    TAGFILTER_TAGS_magic,
    "text/html content of <dl>",
    &(tagfilter_html_dl[0]),
    NULL
};

struct tagfilter_tags text_html_fieldset = {
    TAGFILTER_TAGS_magic,
    "text/html content of <fieldset>",
    &(tagfilter_html_fieldset[0]),
    NULL
};

struct tagfilter_tags text_html_figure = {
    TAGFILTER_TAGS_magic,
    "text/html content of <figure>",
    &(tagfilter_html_figure[0]),
    NULL
};

struct tagfilter_tags text_html_footer = {
    TAGFILTER_TAGS_magic,
    "text/html content of <footer>",
    &(tagfilter_html_footer[0]),
    NULL
};

struct tagfilter_tags text_html_form = {
    TAGFILTER_TAGS_magic,
    "text/html content of <form>",
    &(tagfilter_html_form[0]),
    NULL
};

struct tagfilter_tags text_html_hgroup = {
    TAGFILTER_TAGS_magic,
    "text/html content of <hgroup>",
    &(tagfilter_html_hgroup[0]),
    NULL
};

struct tagfilter_tags text_html_label = {
    TAGFILTER_TAGS_magic,
    "text/html content of <label>",
    &(tagfilter_html_label[0]),
    NULL
};

struct tagfilter_tags text_html_list_items = {
    TAGFILTER_TAGS_magic,
    "text/html list items",
    &(tagfilter_html_list_items[0]),
    NULL
};

struct tagfilter_tags text_html_close_li = {
    TAGFILTER_TAGS_magic,
    "text/html closing tags for <li>",
    &(tagfilter_html_close_li[0]),
    NULL
};

struct tagfilter_tags text_html_close_p = {
    TAGFILTER_TAGS_magic,
    "text/html closing tags for <p>",
    &(tagfilter_html_close_p[0]),
    NULL
};

struct tagfilter_tags text_html_address = {
    TAGFILTER_TAGS_magic,
    "text/html content of <address>",
    &(tagfilter_html_address[0]),
    NULL
};

struct tagfilter_tags text_html_meter = {
    TAGFILTER_TAGS_magic,
    "text/html content of <meter>",
    &(tagfilter_html_meter[0]),
    NULL
};

struct tagfilter_tags text_html_picture = {
    TAGFILTER_TAGS_magic,
    "text/html content of <picture>",
    &(tagfilter_html_picture[0]),
    NULL
};

struct tagfilter_tags text_html_progress = {
    TAGFILTER_TAGS_magic,
    "text/html content of <progress>",
    &(tagfilter_html_progress[0]),
    NULL
};

struct tagfilter_tags text_html_ruby = {
    TAGFILTER_TAGS_magic,
    "text/html content of <ruby>",
    &(tagfilter_html_ruby[0]),
    NULL
};

struct tagfilter_tags text_html_table = {
    TAGFILTER_TAGS_magic,
    "text/html content of <table>",
    &(tagfilter_html_table[0]),
    NULL
};

struct tagfilter_tags text_html_div_dl = {
    TAGFILTER_TAGS_magic,
    "text/html content of <div> on <dl>",
    &(tagfilter_html_div_dl[0]),
    NULL
};

struct tagfilter_tags text_html_caption = {
    TAGFILTER_TAGS_magic,
    "text/html content of <caption>",
    &(tagfilter_html_caption[0]),
    NULL
};

struct tagfilter_tags text_html_colgroup = {
    TAGFILTER_TAGS_magic,
    "text/html content of <colgroup>",
    &(tagfilter_html_colgroup[0]),
    NULL
};

struct tagfilter_tags text_html_tr_list = {
    TAGFILTER_TAGS_magic,
    "text/html <tr> items",
    &(tagfilter_html_tr_list[0]),
    NULL
};

struct tagfilter_tags text_html_tr = {
    TAGFILTER_TAGS_magic,
    "text/html content of <tr>",
    &(tagfilter_html_tr[0]),
    NULL
};

struct tagfilter_tags text_html_close_thead = {
    TAGFILTER_TAGS_magic,
    "text/html closing tags for <thead>",
    &(tagfilter_html_close_thead[0]),
    NULL
};

struct tagfilter_tags text_html_close_tbody = {
    TAGFILTER_TAGS_magic,
    "text/html closing tags for <tbody>",
    &(tagfilter_html_close_tbody[0]),
    NULL
};

struct tagfilter_tags text_html_close_tr = {
    TAGFILTER_TAGS_magic,
    "text/html closing tags for <tr>",
    &(tagfilter_html_close_tr[0]),
    NULL
};

struct tagfilter_tags text_html_close_td = {
    TAGFILTER_TAGS_magic,
    "text/html closing tags for <td>",
    &(tagfilter_html_close_td[0]),
    NULL
};

struct tagfilter_tags text_html_th = {
    TAGFILTER_TAGS_magic,
    "text/html content of <th>",
    &(tagfilter_html_th[0]),
    NULL
};

struct tagfilter_tags text_html_close_th = {
    TAGFILTER_TAGS_magic,
    "text/html closing tags for <th>",
    &(tagfilter_html_close_th[0]),
    NULL
};

struct tagfilter_tags text_html_close_dtdd = {
    TAGFILTER_TAGS_magic,
    "text/html closing tags for <dt> and <dd>",
    &(tagfilter_html_close_dtdd[0]),
    NULL
};

struct tagfilter_tags text_html_close_rtrp = {
    TAGFILTER_TAGS_magic,
    "text/html closing tags for <rt> and <rp>",
    &(tagfilter_html_close_rtrp[0]),
    NULL
};

E_(give_top_stack_item_f give_top_text_html)
struct tagfilter_stack_item * give_top_text_html
  P_((mime_t                     * body,
      struct in_state            * state_in,
      struct out_state           * state_out,
      const struct decode_opts   * decode_opt,
      charset_t                    text_charset,
      struct tagfilter_selection * tagfilter));
struct tagfilter_stack_item * give_top_text_html(body,state_in,state_out,
						 decode_opt,text_charset,
						 tagfilter)
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     const struct decode_opts   * decode_opt;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    struct tagfilter_stack_item * ret = NULL;
    struct pager_range          * html_body_range = NULL;

    /* What is good settings ? */

    
    if (MIME_magic != body->magic)
	mime_panic(__FILE__,__LINE__,"give_top_text_html",
		   "Bad magic number (mime_t)");

    if (decode_opt)
	html_body_range =
	    state_add_simple_pager_range(state_out,
					 /* Handles quote prefix */
					 decode_opt->range,
					 PR_WORD_WRAP|PR_JOIN_LINES|
					 PR_ADD_SPACE,0,0);
    else
	html_body_range =
	    state_add_simple_pager_range(state_out,
					 NULL,
					 PR_WORD_WRAP|PR_JOIN_LINES|
					 PR_ADD_SPACE,0,0);

    ret =  tagfilter_new_stack_item(html_body_range,
				    0,
				    &text_html_top,
				    NULL);

    if (TAGFILTER_STACK_ITEM_magic != ret->magic)
	mime_panic(__FILE__,__LINE__,"give_top_text_html",
		   "Bad magic number (tagfilter_stack_item)");

    free_pager_range(&html_body_range);
    
    return ret;
}

static struct pager_range *new_html_item_hlp1
   P_((struct tagfilter_cmd       * current_tag,
       struct pager_range         * prev_range,
       struct out_state           * state_out,
       int                        * text_visible_p,
       int                          inherit_pg_flags

       ));
static struct pager_range *
    new_html_item_hlp1(current_tag,
		       prev_range,
		       state_out,
		       text_visible_p,
		       inherit_pg_flags)
     struct tagfilter_cmd       * current_tag;
     struct pager_range         * prev_range;
     struct out_state           * state_out;
     int                        * text_visible_p;
     int                          inherit_pg_flags;
{
    struct pager_range * new_range = NULL;
    int  text_visible  = * text_visible_p;

    if (TAGFILTER_CMD_magic != current_tag->magic)
	mime_panic(__FILE__,__LINE__,
		   "new_html_item_hlp1",
		   "Bad magic number (tagfilter_cmd)");

    if (text_visible && ison(current_tag->cmd_mode,
			     TAGFLT_CMD_hide_text))
	text_visible = 0;
    else if (ison(current_tag->cmd_mode,
		  TAGFLT_CMD_show_text))
	text_visible = 1;

    
    switch (current_tag->code.html) {
	int i;
    case tfc_html_unsupported:    
    case tfc_html_none:     break;	
    case tfc_html_br: {
	int prev_flags = 0;
		
	if (prev_range &&
	    (prev_flags = get_pager_range_flags(prev_range)) &&
	    ison(prev_flags,PR_JOIN_LINES)) {

	    /* New range without flags */

	    new_range =
		state_add_simple_pager_range(state_out,
					     prev_range,
					     PR_FLAGS_ABSOLUTE,
					     0,0);


	}
	
	/* Re-set pg_flags */
	set_out_state_line_mode(state_out,inherit_pg_flags,
				new_range ? new_range : prev_range,
				1 /* Newline */);

	state_putc('\n',state_out);

    }
	break;
	
    case tfc_html_hr:
	state_putc('\n',state_out); 

	new_range =
	    state_add_simple_pager_range(
			state_out,
			prev_range,					 	   /* Not very good */
		        PR_CENTER_THIS|PR_FLAGS_ABSOLUTE, 
		        0 /* quote level */,
		        0 /* indent */);
	
	/* Re-set pg_flags */
	set_out_state_line_mode(state_out,inherit_pg_flags,
				new_range,0 );


	for (i = 0; i < 10; i++)
	    state_putc('-',state_out); 

	break;

    case tfc_html_p:
	state_putc('\n',state_out);

	new_range =
	    state_add_simple_pager_range(state_out,
					 prev_range,
					 PR_WORD_WRAP|PR_JOIN_LINES|
					 PR_ADD_SPACE,
					 0 /* quote level */,
					 0 /* indent */);
	
	/* Re-set pg_flags */
	set_out_state_line_mode(state_out,inherit_pg_flags,
				new_range,0 );
	break;

    case tfc_html_pre: {
	/* Calculate absolute range flags */
	int range_flags = PR_FLAGS_ABSOLUTE;
		
	if (prev_range)
	    setit(range_flags,get_pager_range_flags(prev_range));
	clearit(range_flags,PR_JOIN_LINES);
	clearit(range_flags,PR_ADD_SPACE);
	clearit(range_flags,PR_WORD_WRAP);
	clearit(range_flags,PR_SAME_LINE);

	new_range = state_add_simple_pager_range(state_out,
						 prev_range,
						 range_flags,
						 0,0);
	/* Re-set pg_flags */
	set_out_state_line_mode(state_out,inherit_pg_flags,
				new_range,0 );

    }
	break;
    case tfc_html_q:
	state_putc('"',state_out);
	break;
    case tfc_html_li:
	
	new_range = state_add_simple_pager_range(state_out,
						 prev_range,
						 0,
						 0,0);
	set_out_state_line_mode(state_out,inherit_pg_flags,
				new_range,0 );
	break;
       
    }
   
    * text_visible_p = text_visible;

    return new_range;
}

/* XXX Currently dummy -- parameters need check when used */
static void new_html_item_hlp2 P_((struct tagfilter_cmd       * current_tag,
				   struct pager_range         * new_range,
				   struct tagfilter_tag_state * tag_state,
				   struct out_state           * state_out,
				   int                          text_visible,
				   struct tagfilter_selection * tagfilter,
				   struct tagfilter_stack_item * stack_item
				   ));
static void new_html_item_hlp2(current_tag,new_range,
			       tag_state,state_out,
			       text_visible,tagfilter,
			       stack_item)
     struct tagfilter_cmd       * current_tag;
     struct pager_range         * new_range;
     struct tagfilter_tag_state * tag_state;
     struct out_state           * state_out;
     int                          text_visible;
     struct tagfilter_selection * tagfilter;
     struct tagfilter_stack_item * stack_item;
{
    if (TAGFILTER_CMD_magic != current_tag->magic)
	mime_panic(__FILE__,__LINE__,
		   "new_html_item_hlp2",
		   "Bad magic number (tagfilter_cmd)");

    switch (current_tag->code.html) {
    case tfc_html_unsupported:
    case tfc_html_none:         break;
	
	
    case tfc_html_br:           break;
    case tfc_html_hr:           break;
	
    case tfc_html_p:            break;
    case tfc_html_pre:          break;
    case tfc_html_q:            break;
    case tfc_html_li:            break;

	
    }

}



static struct tagfilter_stack_item * process_new_html_item2
  P_((struct tagfilter_cmd       * current_tag,
      struct pager_range         * prev_range,
      struct tagfilter_tag_state * current_state,
      struct out_state           * state_out,
      struct tagfilter_selection * tagfilter,
      int                          is_last,
      struct tagfilter_tags      * used_tags,
      int                          inherit_pg_flags,
      int                          text_visible,
      const struct decode_opts   * decode_opt
      ));
static struct tagfilter_stack_item *
        process_new_html_item2(current_tag,prev_range,
			       current_state,
			       state_out,
			       tagfilter,is_last,
			       used_tags,
			       inherit_pg_flags,
			       text_visible,
			       decode_opt
			       )
     struct tagfilter_cmd       * current_tag;
     struct pager_range         * prev_range;
     struct tagfilter_tag_state * current_state;
     struct out_state           * state_out;
     struct tagfilter_selection * tagfilter;
     int                          is_last;
     struct tagfilter_tags      * used_tags;
     int                          inherit_pg_flags;
     int                          text_visible;
     const struct decode_opts   * decode_opt;
{

    struct tagfilter_stack_item * ret = NULL;
    struct pager_range * new_range = NULL;
    
    if (TAGFILTER_CMD_magic != current_tag->magic)
	mime_panic(__FILE__,__LINE__,"process_new_html_item2",
		   "Bad magic number (tagfilter_cmd)");

    if (current_tag->ascii_command) {
	DPRINT(Debug,21,(&Debug,
			 "process_new_html_item2: command %s%s\n",
			 current_tag->ascii_command,
			 is_last ? ", is last" : ""));
    }
    

    new_range = new_html_item_hlp1(current_tag,
				   prev_range,
				   state_out,
				   &text_visible,
				   inherit_pg_flags);
   
     if (!new_range &&
	 ison(current_tag->cmd_mode,TAGFLT_CMD_force_newline)) {
	 
	 DPRINT(Debug,21,(&Debug,
			  "process_new_html_item2: TAGFLT_CMD_force_newline set, forcing new range\n"));
	 
	 new_range = state_add_simple_pager_range(state_out,
						  prev_range,
						  0,0,0);
	 
     }


     DPRINT(Debug,21,(&Debug,
		      "process_new_html_item2: pg_flags %d %s\n",
		      current_tag->pg_flags,
		      give_pg_flags(current_tag->pg_flags)));

     
    ret = tagfilter_new_stack_item(new_range,
				   current_tag->pg_flags,
				   ison(current_tag->cmd_mode,
					TAGFLT_CMD_inherit_nested_tags) ?
				   used_tags  /* Transparent content model */ :
				   NULL /* Use current_tag */,
				   current_tag);
    
    if (TAGFILTER_STACK_ITEM_magic != ret->magic)
	mime_panic(__FILE__,__LINE__,"process_new_html_item2",
		   "Bad magic number (tagfilter_stack_item)");

    ret->text_visible = text_visible;

    new_html_item_hlp2(current_tag,new_range,
		       current_state,state_out,text_visible,
		       tagfilter,ret);
        
    if (new_range)
	free_pager_range(& new_range);

    return ret;
}

				   
static struct tagfilter_tag_state * process_new_html_item
    P_((struct tagfilter_cmd       * current_tag,
	struct pager_range         * prev_range,
	struct tagfilter_stack_item **new_item,
	struct out_state           * state_out,
	struct tagfilter_global    * counter,
	struct tagfilter_selection * tagfilter,
	struct tagfilter_tags      * tags,
	int                         push_current,
	struct tagfilter_tags      * used_tags,
	int                          inherited_pg_flags,
	int text_visible,
	const struct decode_opts   * decode_opt	
	));
static struct tagfilter_tag_state *
                process_new_html_item(current_tag,
				      prev_range,
				      new_item,
				      state_out,
				      counter,
				      tagfilter,tags,
				      push_current,
				      used_tags,
				      inherit_pg_flags,
				      text_visible,
				      decode_opt
				      )
     struct tagfilter_cmd       * current_tag;
     struct pager_range         * prev_range;
     struct tagfilter_stack_item **new_item;
     struct out_state           * state_out;
     struct tagfilter_global    * counter;
     struct tagfilter_selection * tagfilter;
     struct tagfilter_tags      * tags;
     int                          push_current /* Do not generate new_item now */;
     struct tagfilter_tags      * used_tags;
     int                          inherit_pg_flags;
     int                          text_visible;
     const struct decode_opts   * decode_opt;
{

    struct pager_range * new_range = NULL;
    struct tagfilter_tag_state *ret = NULL;
    
    if (TAGFILTER_CMD_magic != current_tag->magic)
	mime_panic(__FILE__,__LINE__,
		   "process_new_html_item",
		   "Bad magic number (tagfilter_cmd)");

    if (current_tag->ascii_command) {
	DPRINT(Debug,21,(&Debug,
			 "process_new_html_item: command %s%s\n",
			 current_tag->ascii_command,
			 push_current ? ", push current" : ""));
    }
    
    new_range = new_html_item_hlp1(current_tag,
				   prev_range,
				   state_out,
				   &text_visible,
				   inherit_pg_flags);

    if (tfc_html_unsupported == current_tag->code.html) {
	tagfilter_add_unknown_tag(counter,current_tag,
				  tagfilter);
    } else if (tfc_html_li == current_tag->code.html) {
	/* XXX still unsupported */
	
	tagfilter_add_unknown_tag(counter,current_tag,
				  tagfilter);

    }
    
    if (current_tag->code.html > tfc_html_none)
	ret = tagfilter_new_tag_state(current_tag);

     if (!new_range &&
	 ison(current_tag->cmd_mode,TAGFLT_CMD_force_newline)) {
	 
	 DPRINT(Debug,21,(&Debug,
			  "process_new_html_item: TAGFLT_CMD_force_newline set, forcing new range\n"));
	 
	 new_range = state_add_simple_pager_range(state_out,
						  prev_range,
						  0,0,0);
	 
     }
    
    if (new_item && !push_current) {

	DPRINT(Debug,21,(&Debug,
			 "process_new_html_item: pg_flags %d %s\n",
			 current_tag->pg_flags,
			 give_pg_flags(current_tag->pg_flags)));
	
	*new_item = tagfilter_new_stack_item(new_range,
					     current_tag->pg_flags,
					     ison(current_tag->cmd_mode,
						  TAGFLT_CMD_inherit_nested_tags) ?
					     used_tags  /* Transparent content model */ :
					     NULL /* Use current_tag */,
					     current_tag);
	
	if (TAGFILTER_STACK_ITEM_magic != (*new_item)->magic)
	    mime_panic(__FILE__,__LINE__,"process_new_html_item",
		       "Bad magic number (tagfilter_stack_item)");
	
	(*new_item)->text_visible = text_visible;

    } else if (push_current && ret) {
	if (TAGFILTER_TAG_STATE_magic != ret->magic)
	   	mime_panic(__FILE__,__LINE__,"process_new_html_item",
		   "Bad magic number (tagfilter_tag_state)");
 

	
	if (!ret->u.html)
	    ret->u.html = tagfilter_new_html_state();

	tagfilter_html_state_push_command(ret->u.html,
					  current_tag,
					  used_tags);
	
    }

    new_html_item_hlp2(current_tag,new_range,
		       ret,state_out,
		       text_visible,tagfilter,
		       new_item ? *new_item: NULL);
		       

    if (new_range)
	free_pager_range(& new_range);
    	
    return ret;
}

struct tag_search_item {

    struct tagfilter_push_tag  * cmd_list;
    int                          cmd_list_len;
    
    struct tagfilter_tags      * tags;
};

/* Returns found item -- not refcounted */

static struct tag_search_item * generate_search_list
P_((struct tagfilter_tags      * tags,
    struct tag_search_item    ** search_xxx,
    int                        * search_xxx_len,
    struct string              * tag_name,
    struct tagfilter_selection * sel_backlink
    ));
static struct tag_search_item * generate_search_list(tags,
						     search_xxx,
						     search_xxx_len,
						     tag_name,
						     sel_backlink)
     struct tagfilter_tags      * tags;
     struct tag_search_item    ** search_xxx;
     int                        * search_xxx_len;
     struct string              * tag_name;
     struct tagfilter_selection * sel_backlink;
{
    struct tag_search_item    * ret = NULL; 
    struct tag_search_item    * search_list = * search_xxx;
    int                         search_list_len = * search_xxx_len;
    int i;
    
    for (i = -1; i < search_list_len; i++) {
	struct tagfilter_tags * t = NULL;
	struct tagfilter_cmd  * c = NULL;
		
	struct tagfilter_push_tag     * c_list      = NULL;
	int                             c_list_len  = 0;
	
	int k;
	
	if (i < 0) {
	    t = tags;

	    c_list      = NULL;
	    c_list_len  = 0;	
	    
	} else {
	    t           = search_list[i].tags;

	    c_list      = search_list[i].cmd_list;
	    c_list_len  = search_list[i].cmd_list_len;	    
	}

	if (!t)
	    continue;
	
	if (TAGFILTER_TAGS_magic != t->magic)
	    mime_panic(__FILE__,__LINE__,"generate_search_list",
		       "Bad magic number (tagfilter_tags)");

	DPRINT(Debug,21,(&Debug,
			 "generate_search_list: %d: Checking %s for %S\n",
			 i,t->context,tag_name));
			 
	
	c = locate_command(t,tag_name,sel_backlink);

	if (c) {
	    int j;
	    
	    if (TAGFILTER_CMD_magic != c->magic)
		mime_panic(__FILE__,__LINE__,"generate_search_list",
			   "Bad magic number (tagfilter_cmd)");
	    
	    search_list = safe_array_realloc(search_list,
					     search_list_len+1,
					     sizeof (search_list[0]));

	    /* Not really needed */
	    search_list[search_list_len].tags = c->nesting;
	    
	    search_list[search_list_len].cmd_list =
		safe_calloc(c_list_len+1,
			    sizeof (search_list[search_list_len].cmd_list[0]));

	    for (j = 0; j < c_list_len; j++) {
		search_list[search_list_len].cmd_list[j] = c_list[j];

		tagfilter_inc_cmd_refcount(search_list[search_list_len].
					   cmd_list[j].cmd);	
	    }

	    search_list[search_list_len].cmd_list[j].cmd       = c;
	    search_list[search_list_len].cmd_list[j].used_tags = t;
	    tagfilter_inc_cmd_refcount(search_list[search_list_len].
				       cmd_list[j].cmd);
	    search_list[search_list_len].cmd_list_len = j+1;

	    ret = & (search_list[search_list_len] );
	    search_list_len++;
	    goto done;	    
	}

	for (k = 0; t->tags[k]; k++) {
	    int a;
	    int j;
	    struct tagfilter_cmd  * C;
	    
	    if (TAGFILTER_CMD_magic != t->tags[k]->magic)
		mime_panic(__FILE__,__LINE__,"generate_search_list",
			   "Bad magic number (tagfilter_cmd)");

	    C = t->tags[k];
	    
	    if (! C->nesting) 
		continue;

	    

	    /* Only search nested if opening tag is optional */
	    if (isoff(C-> cmd_mode,TAGFLT_CMD_nested_search)) {

		if (C->ascii_command) {

		    DPRINT(Debug,22,(&Debug,
				     "generate_search_list: %d: %d: %s on %s does not support nested search\n",
				     i,k,C->ascii_command,
				     t->context));

		    
		}
		
		continue;
	    }

	    
	    for (a = 0; a < search_list_len; a++) {

		if (C->nesting == search_list[a].tags) {
		    goto found;
		}		    
	    }

	    if (C->ascii_command) {		
		DPRINT(Debug,21,(&Debug,
				 "generate_search_list: %d: %d: %s on %s -> added to search list %d\n",
				 i,k,C->ascii_command,
				 t->context,search_list_len));
	    }

	    search_list = safe_array_realloc(search_list,
					     search_list_len+1,
					     sizeof (search_list[0]));
	    
	    search_list[search_list_len].tags = C->nesting;
	    
	    search_list[search_list_len].cmd_list =
		safe_calloc(c_list_len+1,
			    sizeof (search_list[search_list_len].
				    cmd_list[0]));

	    for (j = 0; j < c_list_len; j++) {
		search_list[search_list_len].cmd_list[j] = c_list[j];

		tagfilter_inc_cmd_refcount(search_list[search_list_len].
					   cmd_list[j].cmd);	
	    }

	    search_list[search_list_len].cmd_list[j].cmd = C;
	    search_list[search_list_len].cmd_list[j].used_tags = t;
	    tagfilter_inc_cmd_refcount(search_list[search_list_len].
				       cmd_list[j].cmd);
	    search_list[search_list_len].cmd_list_len = j+1;

	    search_list_len++;
	    
	found:
	    ;	    
	}	
    }
    
 done:
    
    * search_xxx     = search_list;
    * search_xxx_len = search_list_len;
    
    return ret;
}

/* Invalidates also found item */

static void free_search_list P_((struct tag_search_item ** search_xxx,
				 int                  * search_xxx_len));
static void free_search_list(search_xxx,
			     search_xxx_len)
     struct tag_search_item   ** search_xxx;
     int                       * search_xxx_len;
{
    struct tag_search_item    * search_list     = * search_xxx;
    int                         search_list_len = * search_xxx_len;
    
    if (search_list) {
	int i;

	for (i = 0; i < search_list_len; i++) {

	    if (search_list[i].cmd_list) {
		int j;


		for (j = 0; j < search_list[i].cmd_list_len; j++) {
		    tagfilter_release_cmd(&
					  search_list[i].cmd_list[j].cmd);
		    search_list[i].cmd_list[j].used_tags = NULL;
		}

		free (search_list[i].cmd_list);
		search_list[i].cmd_list = NULL;
	    }
	    search_list[i].cmd_list_len = 0;

	}

	free(search_list);
	search_list = NULL;
    }
    search_list_len = 0;

    * search_xxx     = search_list;
    * search_xxx_len = search_list_len;    
}



E_( give_stack_item_f give_item_text_html)
struct tagfilter_tag_state *
  give_item_text_html P_((struct tagfilter_stack_item **new_item,
			  struct tagfilter_tags      * tags,
			  struct string              * tag_name,
			  struct tagfilter_stack_item **stack,
			  size_t                        stack_len,
			  size_t                     * stack_pos,
			  mime_t                     * body,
			  struct in_state            * state_in,
			  struct out_state           * state_out,
			  const struct decode_opts   * decode_opt,
			  charset_t                    text_charset,
			  struct tagfilter_selection * tagfilter,
			  struct tagfilter_global    * counter
			  ));
struct tagfilter_tag_state *give_item_text_html(new_item,tags,tag_name,
						stack,stack_len,stack_pos,
						body,state_in,state_out,
						decode_opt,
						text_charset, tagfilter,
						counter)
     struct tagfilter_stack_item **new_item;
     struct tagfilter_tags      * tags;
     struct string              * tag_name;
     struct tagfilter_stack_item **stack;
     size_t                        stack_len;
     size_t                     * stack_pos;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     const struct decode_opts   * decode_opt;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
     struct tagfilter_global    * counter;
{
    int text_visible = 1;
    struct tagfilter_tag_state * ret = NULL;
    struct tagfilter_cmd       * current_tag = NULL;

    size_t  STACK_pos = stack_len;

    struct tagfilter_tags      * used_tags = NULL;

    int push_current   = 0;
    
    
    /* Allocates new  tagfilter_tag_state 
       (also NULL return possible) -- 
       gives struct tagfilter_stack_item if
       push_stack_item() needs not be used */
    
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"give_item_text_html",
		   "Bad magic number (tagfilter_selection)");

    if (new_item && *new_item)
	tagfilter_free_stack_item(new_item,tagfilter);

    * stack_pos = stack_len;

    if (!tags && tag_name) {
	DPRINT(Debug,20,(&Debug,
			 "give_item_text_html: %S - no tags, stack_len=%zu\n",
			 tag_name,stack_len));
    } else if (tags && tag_name) {
	if (TAGFILTER_TAGS_magic != tags->magic)
	    mime_panic(__FILE__,__LINE__,"give_item_text_html",
		       "Bad magic number (tagfilter_tags)");
	
	
	DPRINT(Debug,20,(&Debug,
			 "give_item_text_html: %S - for %s, stack_len=%zu\n",
			 tag_name,tags->context,stack_len));
    }
	
    if (tag_name && stack && stack_len > 0) {
	size_t i = stack_len;
       	
	/* Check for auto close */
	
	while (i > 0) {
	    i--;
	    
	    if (stack[i]) {
		if (TAGFILTER_STACK_ITEM_magic != stack[i]->magic)
		    mime_panic(__FILE__,__LINE__,
			       "give_item_text_html",
			       "Bad magic number (tagfilter_stack_item)");
		
		if (! stack[i]->current_tag)
		    continue;
		
		if (TAGFILTER_CMD_magic != stack[i]->current_tag->magic)
		    mime_panic(__FILE__,__LINE__,"give_item_text_html",
			       "Bad magic number (tagfilter_cmd)");
		
		if ( !  stack[i]->current_tag->auto_close)
		    break;
		
		current_tag = locate_command(stack[i]->current_tag->
					     auto_close,
					     tag_name,tagfilter);
		
		if (current_tag) {
		    used_tags = tags;
		    STACK_pos = i;

		    /* Do not generate new_item now */
		    if (stack[i]->current_tag->code.base >= tfc_none)
			push_current = 1;
		    
		}
				
		break;  /* End search anyway */
	    }
	}
    }
	        
    if (tag_name && tags && !current_tag) {

	if (TAGFILTER_TAGS_magic != tags->magic)
	    mime_panic(__FILE__,__LINE__,"give_item_text_html",
		       "Bad magic number (tagfilter_tags)");

	/* locate_command() does not check nested */
	
	current_tag = locate_command(tags,tag_name,tagfilter);
	
	if (current_tag) {
	    used_tags = tags;
	    STACK_pos = stack_len;
	    
	}    
    } 	

    if (tag_name && !current_tag) {

	/* Check upper on on tag */
	
	if (stack && stack_len > 0) {
	    size_t i = stack_len;

	    int maybe_push_current = 0;
	    
	    while (i > 0) {
		i--;
		
		if (stack[i]) {
		    if (TAGFILTER_STACK_ITEM_magic != stack[i]->magic)
			mime_panic(__FILE__,__LINE__,
				   "give_item_text_html",
				   "Bad magic number (tagfilter_stack_item)");
		    
		    if (! stack[i]->current_tag)
			continue;
		    
		    if (TAGFILTER_CMD_magic !=
			stack[i]->current_tag->magic)
			mime_panic(__FILE__,__LINE__,
				   "give_item_text_html",
				   "Bad magic number (tagfilter_cmd)");

		    /* Do not generate new_item now */
		    if (stack[i]->current_tag->code.base >= tfc_none)
			maybe_push_current = 1;
		    
		    
		    if (stack[i]->nested_tags) {
			current_tag = locate_command(stack[i]->
						     nested_tags,
						     tag_name,tagfilter);
			
			if (current_tag) {
			    /* Potentially replaces next tag on stack */
			    
			    STACK_pos = i+1;
			    used_tags = stack[i]->nested_tags;

			    /* Do not generate new_item now */
			    if (maybe_push_current)
				push_current = 1;

			    
			    break;
			}
		    }
		    
		    /* Stop searching if
		       TAGFLT_CMD_close_search
		       in not set */
		    
		    if (isoff(stack[i]->current_tag->cmd_mode,
			      TAGFLT_CMD_close_search))
			break;
		}			    
	    }
	}		
    }

    if (tag_name && !current_tag) {
	struct tag_search_item * search_list = NULL;
	int                      search_list_len = 0;
		
	struct tag_search_item * found_xx
	    = generate_search_list(tags,
				   &search_list,&search_list_len,
				   tag_name,
				   tagfilter);
	
	if (found_xx &&
	    found_xx->cmd_list &&
	    found_xx->cmd_list_len > 0) {
	    int i;
	    
	    ret =
		tagfilter_new_tag_state(found_xx->
					cmd_list[found_xx->
						 cmd_list_len-1].cmd);

	    if (TAGFILTER_TAG_STATE_magic != ret->magic)
	   	mime_panic(__FILE__,__LINE__,"give_item_text_html",
			   "Bad magic number (tagfilter_tag_state)");
	    

	    ret->u.html = tagfilter_new_html_state();

	    if (TAGFILTER_HTML_STATE_magic != ret->u.html->magic) 
		mime_panic(__FILE__,__LINE__,"give_item_text_html",
			   "Bad magic number (tagfilter_selection)");
		
	    ret->u.html->push_tag =
		safe_calloc(found_xx->cmd_list_len,
			    sizeof (ret->u.html->push_tag[0]));
	    
	    for (i = 0; i < found_xx->cmd_list_len; i++) {
		ret->u.html->push_tag[i] = found_xx->cmd_list[i];

		tagfilter_inc_cmd_refcount(ret->u.html->push_tag[i].cmd);
	    }
	    ret->u.html->push_tag_len = found_xx->cmd_list_len;
	} else {
	    DPRINT(Debug,21,(&Debug,
			     "give_item_text_html: %S - not found with nested search - search list len %d\n",
			     tag_name,search_list_len));
	}
	
	free_search_list(&search_list,&search_list_len);		
    }
		         
    if (current_tag && !ret) {
	int                  inherit_pg_flags = 0;
	struct pager_range * prev_range =
	    give_prev_range(stack,
			    STACK_pos,
			    & inherit_pg_flags,
			    & text_visible);
       	
	
	ret = process_new_html_item(current_tag,prev_range,new_item,
				    state_out,counter,
				    tagfilter,used_tags,push_current,
				    used_tags,
				    inherit_pg_flags,
				    text_visible,
				    decode_opt);	        		
    }
    
    if (!current_tag && tag_name && !ret) {

	struct pager_range * prev_range UNUSED_VAROK =
	    give_prev_range(stack,
			    STACK_pos,
			    NULL,
			    & text_visible);

	
	if (counter) 	   
	    current_tag = tagfilter_give_unknown_tag(counter,tag_name,
						     tagfilter,
						     tfc_unsupported,
						     TAGFLT_CMD_inherit_nested_tags);
						     

	if (!current_tag)
	    current_tag = tagfilter_cmd_from_string(tag_name,
						    tfc_unsupported,
						    TAGFLT_CMD_inherit_nested_tags);


	if (TAGFILTER_CMD_magic != current_tag->magic)
	    mime_panic(__FILE__,__LINE__,"give_item_text_html",
		       "Bad magic number (tagfilter_cmd)");

	if (new_item) {

	    DPRINT(Debug,21,(&Debug,
			     "give_item_text_html: %S pg_flags %d %s\n",
			     tag_name,current_tag->pg_flags,
			     give_pg_flags(current_tag->pg_flags)));
	    
	    *new_item = tagfilter_new_stack_item(NULL,
						 current_tag->pg_flags,
						 ison(current_tag->cmd_mode,
						      TAGFLT_CMD_inherit_nested_tags) ?
						 tags  /* Transparent content model */ :
						 NULL /* Use current_tag */,
						 current_tag);
	}
	
	if (TAGFILTER_STACK_ITEM_magic != (*new_item)->magic)
	    mime_panic(__FILE__,__LINE__,"give_item_text_html",
		       "Bad magic number (tagfilter_stack_item)");
	
	
	(*new_item)->text_visible = text_visible;	
    }

    
    * stack_pos = STACK_pos;

    if (current_tag && !ret)
	ret = tagfilter_new_tag_state(current_tag);
    
    if (new_item && *new_item && current_tag) {
	if (TAGFILTER_STACK_ITEM_magic != (*new_item)->magic)
	    mime_panic(__FILE__,__LINE__,"give_item_text_html",
		       "Bad magic number ( tagfilter_stack_item)");
	if (TAGFILTER_CMD_magic != current_tag->magic)
	    mime_panic(__FILE__,__LINE__,"give_item_text_html",
		       "Bad magic number (tagfilter_cmd)");

	(*new_item)->force_eoln =
	    ison(current_tag->cmd_mode,TAGFLT_CMD_force_newline);
    }
   
    
    if (current_tag)
	tagfilter_release_cmd(& current_tag);
    
    return ret;

}

E_(locate_stack_item_f locate_item_text_html)
struct tagfilter_tag_state *
        locate_item_text_html P_((struct string               * tag_name,
				  struct tagfilter_stack_item **stack,
				  size_t                        stack_len,
				  size_t                      * index
				  /* or stack_len if not found */,
				  mime_t                     * body,
				  struct in_state            * state_in,
				  struct out_state           * state_out,
				  charset_t                    text_charset,
				  struct tagfilter_selection * tagfilter));
struct tagfilter_tag_state * locate_item_text_html(tag_name,
						   stack,stack_len,
						   index,body,
						   state_in,state_out,
						   text_charset,tagfilter)
     struct string               * tag_name;
     struct tagfilter_stack_item **stack;
     size_t                        stack_len;
     size_t                      * index
     /* or stack_len if not found */;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    /* Allocates new  tagfilter_tag_state 
       (also NULL return possible) */
    
    struct tagfilter_tag_state * ret = NULL;

   struct string_sort  * tag_name_X = NULL;  /* Possible lowercased
						already 
					     */
   
   if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"locate_item_text_html",
		   "Bad magic number (tagfilter_selection)");
   
   ret = tagfilter_locate_stack_item(tagfilter,
				     &tag_name_X,
				     tag_name,
				     stack,stack_len,index);
   if (ret) {
       if (TAGFILTER_TAG_STATE_magic != ret->magic)
	   mime_panic(__FILE__,__LINE__,"locate_item_text_html",
		      "Bad magic number (tagfilter_tag_state)");

       
       if (! ret->u.html)
	   ret->u.html = tagfilter_new_html_state();
       
       if (TAGFILTER_HTML_STATE_magic != ret->u.html->magic) 
	   mime_panic(__FILE__,__LINE__,"locate_item_text_html",
		      "Bad magic number (tagfilter_selection)");
   }
   
   
   if (tag_name_X)
       free_string_sort(& tag_name_X);
    
   return ret;
}

E_(end_tag_state_f end_tag_text_html)
void end_tag_text_html P_((struct tagfilter_tag_state **tag_state,
			       mime_t                     * body,
			       struct in_state            * state_in,
			       struct out_state           * state_out,
			       charset_t                    text_charset,
			       struct tagfilter_selection * tagfilter));
void end_tag_text_html(tag_state,body,state_in,state_out,text_charset,tagfilter)
     struct tagfilter_tag_state **tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"end_tag_text_html",
		   "Bad magic number (tagfilter_selection)");
    
    
    if (TAGFILTER_TAG_STATE_magic != (*tag_state)->magic)
	mime_panic(__FILE__,__LINE__,"end_tag_text_html",
		   "Bad magic number (tagfilter_tag_state)");

    if ((*tag_state)->u.html)
	tagfilter_free_html_state(& ((*tag_state)->u.html));
    
    tagfilter_free_tag_state_common (tag_state);
}

E_(push_stack_item_f push_tag_text_html)
struct tagfilter_stack_item *
   push_tag_text_html P_((struct tagfilter_tag_state *tag_state,
			  struct tagfilter_stack_item **stack,
			  size_t                        stack_len,
			  mime_t                     * body,
			  struct in_state            * state_in,
			  struct out_state           * state_out,
			  const struct decode_opts   * decode_opt,
			  charset_t                    text_charset,
			  struct tagfilter_selection * tagfilter));
struct tagfilter_stack_item *push_tag_text_html(tag_state,stack,
						stack_len,body,state_in,
						state_out,decode_opt,text_charset,
						tagfilter)
     struct tagfilter_tag_state *tag_state;
     struct tagfilter_stack_item **stack;
     size_t                        stack_len;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     const struct decode_opts   * decode_opt;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    struct tagfilter_stack_item *ret = NULL;
    
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"push_tag_text_html",
		   "Bad magic number (tagfilter_selection)");
    
    
    if (TAGFILTER_TAG_STATE_magic != tag_state->magic)
	mime_panic(__FILE__,__LINE__,"push_tag_text_html",
		   "Bad magic number (tagfilter_tag_state)");

    if (tag_state->u.html) {
	struct tagfilter_cmd    *cmd = NULL;
	struct tagfilter_tags  * used_tags = NULL;
	int is_last = 0;
	
	if (TAGFILTER_HTML_STATE_magic != tag_state->u.html->magic) 
	    mime_panic(__FILE__,__LINE__,"push_tag_text_html",
		      "Bad magic number (tagfilter_selection)");

	if (tag_state->u.html->push_tag) {
	    int i,t = 0;

	    /* This should select first comamnd, and 
	       move others
	    */
	    
	    for (i = 0; i < tag_state->u.html->push_tag_len; i++) {

		if (tag_state->u.html->push_tag[i].cmd) {

		    if (cmd) {

			if (t < i) {
			    if (tag_state->u.html->push_tag[t].cmd ||
				tag_state->u.html->push_tag[t].used_tags)
				mime_panic(__FILE__,__LINE__,
					   "push_tag_text_html",
					   "Target not empty");
			    else {
				tag_state->u.html->push_tag[t++] =
				    tag_state->u.html->push_tag[i];

				tag_state->u.html->push_tag[i].cmd = NULL;
				tag_state->u.html->push_tag[i].used_tags = NULL;
			    }			    
			}

			is_last = 0;

		    } else {
			cmd       = tag_state->u.html->push_tag[i].cmd;
			used_tags = tag_state->u.html->push_tag[i].used_tags;
			
			tag_state->u.html->push_tag[i].cmd = NULL;
			tag_state->u.html->push_tag[i].used_tags = NULL;
			
			is_last = 1;
		    }		    
		}
	    }
	    tag_state->u.html->push_tag_len = t;

	    if (0 == tag_state->u.html->push_tag_len) {
		free(tag_state->u.html->push_tag);
		tag_state->u.html->push_tag = NULL;
	    }	    
	}
	
	if (cmd) {
	    int  inherit_pg_flags = 0;
	    int text_visible = 1;
	    struct pager_range * prev_range =
		give_prev_range (stack,stack_len,
				 & inherit_pg_flags,
				 & text_visible
				 );

	    ret = process_new_html_item2(cmd,prev_range,
					 tag_state,
					 state_out,
					 tagfilter,
					 is_last,
					 used_tags,
					 inherit_pg_flags,
					 text_visible,
					 decode_opt);
	    
	    tagfilter_release_cmd(& cmd);
	}
	    
    }	

    return ret;
}

E_(close_stack_item_f close_tag_text_html)
int close_tag_text_html P_((struct tagfilter_tag_state *tag_state,
			    struct tagfilter_stack_item **stack,
			    size_t                        stack_len,
			    size_t                        idx,
			    enum stack_close_mode         close_pos,
			    mime_t                     * body,
			    struct in_state            * state_in,
			    struct out_state           * state_out,
			    charset_t                    text_charset,
			    struct tagfilter_selection * tagfilter));
int close_tag_text_html(tag_state,stack,stack_len,idx,close_pos,
			    body,state_in,state_out,text_charset,
			    tagfilter)
     struct tagfilter_tag_state *tag_state;
     struct tagfilter_stack_item **stack;
     size_t                        stack_len;
     size_t                        idx;
     enum stack_close_mode         close_pos;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    int ret = 1;   /* Return 1 when item can be emoved from stack */
    struct tagfilter_cmd       * current_tag = NULL;


    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"close_tag_text_html",
		   "Bad magic number (tagfilter_selection)");
    
    if (TAGFILTER_TAG_STATE_magic != tag_state->magic)
	mime_panic(__FILE__,__LINE__,"close_tag_text_html",
		   "Bad magic number (tagfilter_tag_state)");

    if (! stack || idx >= stack_len || ! stack[idx]) 
	mime_panic(__FILE__,__LINE__,"close_tag_text_html",
		   "Bad stack idx or stack item or no stack");
    
    if (TAGFILTER_STACK_ITEM_magic != stack[idx]->magic)
	mime_panic(__FILE__,__LINE__,
		   "close_tag_text_html",
		   "Bad magic number (tagfilter_stack_item)");

    if (! stack[idx]->current_tag)
	goto out;
    
    if (TAGFILTER_CMD_magic != stack[idx]->current_tag->magic)
	mime_panic(__FILE__,__LINE__,"close_tag_text_html",
		   "Bad magic number (tagfilter_cmd)");

    
    current_tag = stack[idx]->current_tag;
    
   
    if (tag_state->u.html) {

   	if (TAGFILTER_HTML_STATE_magic != tag_state->u.html->magic) 
	    mime_panic(__FILE__,__LINE__,"close_tag_text_html",
		      "Bad magic number (tagfilter_selection)");

	switch (current_tag->code.html) {
	case tfc_html_unsupported:    /* Should not happen */
	case tfc_html_none:     break;	

	case tfc_html_br: /* Nothing to do */ break;

	case tfc_html_hr:
	case tfc_html_p:
	    state_putc('\n',state_out); 
	    
	    /* tagfilter_handle_tag() will call set_out_state_line_mode() */
	    break;
	case tfc_html_pre:              break;
	case tfc_html_q:
	    state_putc('"',state_out);  
	    break;
	case tfc_html_li:
	    state_putc('\n',state_out); 
	    
	    /* tagfilter_handle_tag() will call set_out_state_line_mode() */
	    break;	    
	}
    }
    
 out:
	
    return ret;
}

E_(start_attribute_f start_atr_text_html)
struct  tagfilter_atr_state *
   start_atr_text_html P_((struct string              * atr_name,
			 struct tagfilter_tag_state * tag_state,
			 mime_t                     * body,
			 struct in_state            * state_in,
			 struct out_state           * state_out,
			 charset_t                    text_charset,
			 struct tagfilter_selection * tagfilter));
struct  tagfilter_atr_state *start_atr_text_html(atr_name,tag_state,
						     body,state_in,
						     state_out,
						     text_charset,
						     tagfilter)
     struct string              * atr_name;
     struct tagfilter_tag_state * tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{
    
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"start_atr_text_html",
		   "Bad magic number (tagfilter_selection)");

    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"start_atr_text_html",
		   "Bad magic number (tagfilter_tag_state)");

    /* XXX Need code for handle text/html attributes */
        
    return NULL;
}


E_(end_attribute_f end_atr_text_html)
void end_atr_text_html P_((struct tagfilter_atr_state **atr_state,
			       struct tagfilter_tag_state * tag_state,
			       mime_t                     * body,
			       struct in_state            * state_in,
			       struct out_state           * state_out,
			       charset_t                    text_charset,
			       struct tagfilter_selection * tagfilter));
void end_atr_text_html(atr_state,tag_state,body,state_in,state_out,
			   text_charset,tagfilter)
     struct tagfilter_atr_state **atr_state;
     struct tagfilter_tag_state * tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{

    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"end_atr_text_html",
		   "Bad magic number (tagfilter_selection)");

    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"end_atr_text_html",
		   "Bad magic number (tagfilter_tag_state)");

    /* XXX Need code for handle text/html attributes */    
}

E_(start_end_atrvalue_f start_end_atrval_text_html)
void start_end_atrval_text_html
         P_((struct tagfilter_atr_state * atr_state,
	     enum atrvalue_start_end      start_end,
	     struct tagfilter_tag_state * tag_state,
	     mime_t                     * body,
	     struct in_state            * state_in,
	     struct out_state           * state_out,
	     charset_t                    text_charset,
	     struct tagfilter_selection * tagfilter));
void start_end_atrval_text_html(atr_state,start_end,tag_state,
				    body,state_in,state_out,
				    text_charset,tagfilter)
     struct tagfilter_atr_state * atr_state;
     enum atrvalue_start_end      start_end;
     struct tagfilter_tag_state * tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{

    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"start_end_atrval_text_html",
		   "Bad magic number (tagfilter_selection)");

    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"start_end_atrval_text_html",
		   "Bad magic number (tagfilter_tag_state)");

    /* XXX Need code for handle text/html attributes */    

}

E_(atrvalue_seg_string_f atrvalue_seg_str_text_html)
void atrvalue_seg_str_text_html
   P_((struct tagfilter_atr_state * atr_state,
       struct string              * atr_value_segment,
       struct tagfilter_tag_state * tag_state,
       mime_t                     * body,
       struct in_state            * state_in,
       struct out_state           * state_out,
       charset_t                    text_charset,
       struct tagfilter_selection * tagfilter));
void atrvalue_seg_str_text_html(atr_state,atr_value_segment,
				    tag_state,body,state_in,
				    state_out,text_charset,tagfilter)
     struct tagfilter_atr_state * atr_state;
     struct string              * atr_value_segment;
     struct tagfilter_tag_state * tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{

    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"atrvalue_seg_str_text_html",
		   "Bad magic number (tagfilter_selection)");

    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"atrvalue_seg_str_text_html",
		   "Bad magic number (tagfilter_tag_state)");

    /* XXX Need code for handle text/html attributes */
    
}

E_(atrvalue_seg_ent_f atrvalue_seg_ent_text_html)
void atrvalue_seg_ent_text_html
       P_((struct tagfilter_atr_state * atr_state,
	   struct out_entity          * atr_value_segment,
	   struct tagfilter_tag_state * tag_state,
	   mime_t                     * body,
	   struct in_state            * state_in,
	   struct out_state           * state_out,
	   charset_t                    text_charset,
	   struct tagfilter_selection * tagfilter));
void atrvalue_seg_ent_text_html(atr_state,atr_value_segment,
				    tag_state,body,state_in,state_out,
				    text_charset,tagfilter)
     struct tagfilter_atr_state * atr_state;
     struct out_entity          * atr_value_segment;
     struct tagfilter_tag_state * tag_state;
     mime_t                     * body;
     struct in_state            * state_in;
     struct out_state           * state_out;
     charset_t                    text_charset;
     struct tagfilter_selection * tagfilter;
{

    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"atrvalue_seg_ent_text_html",
		   "Bad magic number (tagfilter_selection)");

    if (TAGFILTER_TAG_STATE_magic !=  tag_state->magic)
	mime_panic(__FILE__,__LINE__,"atrvalue_seg_ent_text_html",
		   "Bad magic number (tagfilter_tag_state)");

    /* XXX Need code for handle text/html attributes */

}

E_(stack_item_clear_f stack_item_clear_text_html)
void stack_item_clear_text_html
  P_((struct tagfilter_stack_item * item,
      struct tagfilter_selection  * tagfilter));
void stack_item_clear_text_html(item,tagfilter)
     struct tagfilter_stack_item * item;
     struct tagfilter_selection  * tagfilter;
{
    if (TAGFILTER_STACK_ITEM_magic != item->magic)
	mime_panic(__FILE__,__LINE__,
		   "stack_item_clear_text_html",
		   "Bad magic number (tagfilter_stack_item)");
    
    if (TAGFILTER_SELECTION_magic != tagfilter->magic) 
	mime_panic(__FILE__,__LINE__,"stack_item_clear_text_html",
		   "Bad magic number (tagfilter_selection)");

}

/* tagfilter_cmd --------------------------------------- */

/* Derement refcount -- free if not static */
void tagfilter_release_cmd(cmd)
     struct tagfilter_cmd **cmd;
{
    if (TAGFILTER_CMD_magic != (*cmd)->magic)
	mime_panic(__FILE__,__LINE__,"tagfilter_release_cmd",
		   "Bad magic number (tagfilter_cmd)");
    
    if ((*cmd)->refcount < 1)
	mime_panic(__FILE__,__LINE__,"tagfilter_release_cmd",
		   "Bad refcount");

    (*cmd)->refcount--;

    if ((*cmd)->refcount > 0) {
	*cmd = NULL;
	return;
    }

    /* string_command is always dynamically allocated */

    if ((*cmd)->string_command)
	free_string_sort(& ((*cmd)->string_command));


    /* Rest may be static */
    
    if (ison((*cmd)->cmd_mode,TAGFLT_CMD_is_static)) {
	*cmd = NULL;
	return;
    }

    if ((*cmd)->ascii_command) {
	free((*cmd)->ascii_command);
	(*cmd)->ascii_command = NULL;
    }

    (*cmd)->nesting = NULL;   /* static */

    (*cmd)->magic = 0;
    free(*cmd);
    *cmd = NULL;
}

void tagfilter_inc_cmd_refcount(cmd)
     struct tagfilter_cmd *cmd;
{
    if (TAGFILTER_CMD_magic != cmd->magic)
	mime_panic(__FILE__,__LINE__,"tagfilter_inc_cmd_refcount",
		   "Bad magic number (tagfilter_cmd)");
    cmd->refcount++;
}

void tagfilter_free_sorted_tags()
{
    static struct tagfilter_tags * check_tags [] = {

	& text_enriched_tags,
	& text_html_top,
	& text_html_head_body,
	& text_html_metadata,
	& text_html_flow,
	& text_html_phrasing,
	
	& text_html_close_head,
	& text_html_noscript_head,
	& text_html_datalist,
	& text_html_details,
	& text_html_dl,
	& text_html_fieldset,
	& text_html_figure,
	& text_html_footer,
	& text_html_form,
	& text_html_hgroup,
	& text_html_label,
	& text_html_list_items,
	& text_html_close_li,
	& text_html_close_p,
	& text_html_address,
	& text_html_meter,
	& text_html_picture,
	& text_html_progress,
	& text_html_ruby,
	& text_html_table,
	& text_html_div_dl,
	& text_html_caption,
	& text_html_colgroup,
	& text_html_tr_list,
	& text_html_tr,
	& text_html_close_thead,
	& text_html_close_tbody,
	& text_html_close_tr,
	& text_html_close_td,
	& text_html_th,
	& text_html_close_th,
	& text_html_close_dtdd,
	& text_html_close_rtrp
    };
    static size_t check_tags_count = (sizeof check_tags) / sizeof (check_tags [0]);
    
    size_t x;

    for (x = 0; x < check_tags_count; x++) {
	if (check_tags[x]) {
	    if (TAGFILTER_TAGS_magic != check_tags[x]->magic) 
		mime_panic(__FILE__,__LINE__,"tagfilter_free_sorted_tags",
			   "Bad magic number (tagfilter_tags)");
	    
	    if (check_tags[x]->sorted_tags)
		free_sort_list( &(check_tags[x]->sorted_tags));
		
	}
    }

}



/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */

